;; -*- package: belief-specialist -*-

;; This specialist implements simulative reasoning about belief, as
;; documented in Aaron Kaplan's PhD thesis (2000, University of
;; Rochester).

;; Known bugs:

;; There is one simulation environment for each term that names a
;; believer.  If two such terms are equated, the contents of their
;; simulation environments should be combined, and a single
;; environment used for both of them thenceforth.   This isn't done.

;; I'm not confident about the interaction between subnets and
;; simulations.  Ultimately, I think we should probably settle on
;; either subnets or environments for storing modally embedded
;; propositions, because having both around is too confusing.  Subnets
;; within simulations should work, but simulations in subnets
;; don't.  E.g. (john believe (that (mary say (that (lrrh girl))))) is
;; ok, but switch the verbs and it's not.


(in-package user)
(eval-when (compile load eval)
  (provide 'belief-specialist)
  (require 'epilib "epilog:epilib;epilib")
  (require 'spec "epilog:spec;spec")
  (defpackage belief-specialist
    (:nicknames belief)
    (:use lisp user spec lib epilib)
    (:export enter evaluate *shared-kb* *at-top-level* skn simulation-env))
  (unless (find-package 'belief) (make-package 'belief))
)

(in-package belief)

;;; We define *skb*, the shared knowledge base, as an environment
;;; accessible to the system itself as well as to simulations.  It
;;; stores common knowledge.

(defvar *shared-kb* (make-environment :name "Shared world knowledge"))
(push *shared-kb* (environment-use-environments *environment*))
(defvar *at-top-level* t "nil during a simulation, t otherwise")

(traceable 'epilog::belief-enter
	   "Traces entry of literals into the belief specialist")
(traceable 'epilog::belief-evaluate
	   "Traces evaluation of literals by the belief specialist")
(traceable 'epilog::belief-details
	   "Traces details of belief specialist operation.")
(traceable 'epilog::belief-all
	   "Traces everything about the belief specialist"
	   '(epilog::belief-enter epilog::belief-evaluate epilog::belief-details))

(defvar *simulation-depth* 0)
(tweakable 'epilog::*max-simulation-depth* 3
	   "The maximum allowed embedding of a simulation of a
simulation of ..." 'integer)
(tweakable 'epilog::*simulation-effort-level* 3
	   "The effort level used in simulative queries")

(defun simulation-env (believer)
  (geth believer
	(retrieve-specialist-info *environment*
				  'belief
				  #'make-hash-table)))

(defun set-simulation-env (believer env)
  (setf (geth believer
	      (retrieve-specialist-info *environment*
					'belief
					#'make-hash-table))
    env))

(defsetf simulation-env set-simulation-env)

(defun approve-believer (term lit)
  (cond ((or (eq term 'epilog::me)
	     (epi-variable-p term))
	 (if (traced-p 'epilog::belief-details)
	     (print-line "Belief specialist: doing nothing with " lit
			 " because the believer term " term
			 " is not one I handle."))
	 nil)
	(t)))

(defun enter (netname lit negp pred arglist &aux (believer (car arglist)))
  (unless (approve-believer believer lit)
    (return-from enter 'unknown))

  (assert (eq pred 'epilog::believe))

  (when (>= *simulation-depth* epilog::*max-simulation-depth*)
    (if (traced-p 'epilog::belief-details)
	(print-line "Belief specialist: not entering " lit
		    " into simulation because maximum nesting exceeded."))
    (return-from enter nil))
  (if (not (equal netname '/))
      (break "tried to enter something into the belief specialist within a subnet"))
  (if negp
      nil
    (let ((sim (simulation-env believer))
	  (belief-content (belief-content (cadr arglist))))
      
      (when (null sim)
	(setq sim (make-environment :name (concatenate 'string
					    "Beliefs of "
					    (string believer))
				    :use-environments (list *shared-kb*)))
	(setf (simulation-env believer)
	  sim))				; don't make this
					; step rewindable, otherwise
					; you end up with agents with
					; null simulations
      (when (null belief-content)
	(if (traced-p 'epilog::belief-enter)
	    (print-line "Belief specialist: don't know how to handle belief wff "
			lit))
	(return-from enter nil))

      ;; This point is reached only if we were able to find the
      ;; content of the belief.

      ;; belief is now a wff name (a symbol).  We can't use the same
      ;; wff name in the simulation, because the symbol's property
      ;; list contains information like input-index and parents that
      ;; can't be shared between the simulation and the real thing.  So get
      ;; its print name, and enter it into the simulation from scratch.
      ;; Eventually, maybe I should pull that info out of the plist
      ;; and put it somewhere simulation-specific, and then reuse the
      ;; wff symbol to avoid the wasted effort of re-normalizing.
      
      (when (traced-p 'epilog::belief-enter)
	(print-line "Belief specialist: beginning simulation of "
		    believer " adding belief " belief-content))
		    
      (let ((result (simulate-in-environment
		     sim
		     `(epilog:story ',belief-content))))
	(when (traced-p 'epilog::belief-enter)
	  (print-line "Belief specialist: simulation returned " result))
	;; If the simulation made any input-driven inferences, embed
	;; each one in a belief sentence and add that to the
	;; *assertions* list.
	(setq *assertions*
	  (append *assertions*
		  (mapcar #'(lambda (belief-name)
			      (make-spec-inference
			       (list believer
				     'epilog::believe
				     (list 'epilog::that (getp belief-name
							       'print)))))
			  ;; I think the first element of the list is
			  ;; always the one that was originally
			  ;; asserted, so don't use that.
			  (cdr result))))
	(if result t nil)))))

(defun evaluate (netname lit negp pred arglist effort &aux (believer
							    (car arglist)))
  (declare (ignore effort))

  (if (not (approve-believer believer lit))
      (return-from evaluate 'unknown))
  
  (assert (eq pred 'epilog::believe))

  (when (>= *simulation-depth* epilog::*max-simulation-depth*)
    (if (traced-p 'epilog::belief-details)
	(print-line "Belief specialist: not evaluating " lit
		    " in simulation because maximum nesting exceeded."))
    (return-from evaluate 'unknown))
  (if (not (equal netname '/))
      (break "tried to enter something into the belief specialist within a subnet"))
  (let* ((believer (car arglist))
	 (sim (simulation-env believer))
	 (belief-content (belief-content (cadr arglist))))
    (when (null sim)
      (setq sim (make-environment :name (concatenate 'string
					  "Beliefs of "
					  (string believer))
				  :use-environments (list *shared-kb*)))
      (setf (simulation-env believer) sim))
					; don't make this
					; step rewindable, otherwise
					; you end up with agents with
					; null simulations
    (when (null belief-content)
      (if (traced-p 'epilog::belief-enter)
 	  (print-line "Belief specialist: don't know how to handle
 		 belief wff " lit))
      (return-from evaluate 'unknown))

    ;; This point is reached only if we were able to find the
    ;; content of the belief.

    (when (traced-p 'epilog::belief-evaluate)
      (print-line "Belief specialist: beginning simulation of " believer
		  " with query " belief-content))
    (let ((result (caar (simulate-in-environment
			 sim
			 `(prog1
			      (epilog:question
			       ',belief-content
			       1
			       ,epilog::*simulation-effort-level*)
			    (epilog:rewind-question))))))
      ;; If q returns multiple answers, we only use the first.
      (when (traced-p 'epilog::belief-evaluate)
	(print-line "Belief specialist: simulation answered " result))
      (if negp
	  (case result
	    ((nil) 'unknown)
	    (yes 'no)
	    (no 'yes))
	(case result
	  ((nil) 'unknown)
	  (yes 'yes)
	  (no 'no))))))


;; term is either a propositional constant, or a list of the form
;; (that S), where S is a wff.  Return the name of the content wff, or
;; nil if we can't find the content.  A propositional constant
;; shouldn't have more than one content sentence; if it does, we just use
;; the first one.
(defun belief-content (term)
  (cond ((and (eq (getp term 'term-type) 'modified)
	      (eq (getp term 'operator) 'that))
	 (getp (car (getp term 'arguments)) 'print))
	((and (eq (getp term 'term-type) 'constant)
	      (eq (getp term 'sort) 'propos))
	 (car (content term)))
	(t nil)))



(in-package epilib)

;; Evaluate form in the simulation environment instead of the "real
;; thing," introducing a new dynamic scope for any variable that might
;; be modified during an assertion or query.  Returns the return value
;; of the last form.

;; note that this defun needs to be evaluated in the epilib
;; environment to get all the variables from the right packages.  It
;; would be less error-prone to specify each variable's package
;; explicitly.
(defun belief::simulate-in-environment (env form)
  (let (
	(*environment* env)
	(belief::*simulation-depth* (1+ belief::*simulation-depth*))
	(belief:*at-top-level* nil)

	;; special variables of which we don't want a
	;; simulation to change its parent's value, but for
	;; which values needn't be persistent across
	;; invocations of a simulation.  Values are the same
	;; given in the respective defvar declarations.
	epilib::*typed-variables*
	*variables*
	*current-arg*
	*lambda-arg*
	*embedded*
	*main-class*
	*reverse-test*
	*subs*
	*constants-to-update*
	(*temp-table*
	 (initialize-element-table
	  "Temporary table for formulas and subformulas"))
	*perm-memory-output*
	*perm-items*
	*perm-probs*
	*simplification-schemas*
	*interim*
	(*normed-prob* 1)
	*normed-subs*
	*skol-subs*
	*wh-vars*
	*e-vars*
	(*new-vars* t)
	*got-sort*
	*quoted*
	*cur-var*
	*cur-varnum*
	*vars-to-fix*
	*vars-to-restore*
	*+lits*
	*-lits*
	*cur-var-matchable*
	*cur-var-nonmatchable*
	(*cur-varnum-matchable* 0)
	(*cur-varnum-nonmatchable* 0)
	*prob-already*
	*recalced-wffs*
	said
	*multiple-subs*
	(*subst-prob* 1)
	*uf-used-indices*
	*uf-functions*
	*assertions*
	part::*current-part-literal*
	spec::*current-literal*
	spec::*flattened-lit*
	spec::*changed-concepts*
	epilog::*save-class*
	epilog::*save-intermediate*
	epilog::*intermediate-class*
	epilog::*meaning-postulate*
	epilog::*simplification-schema*
	epilog::*meta*
	epilog::*goal-only*
	epilog::*pieces*
	epilog::*ordered-pieces*
	epilog::*exists-q*
	epilog::*do-not-enter-spec*
	epilog::*key-verify*
	epilog::*ok-to-verify*
	epilog::*old-qa-effort*
	epilog::*qa-mode*
	(epilog::*qa-start-time* 0)
	(epilog::*qa-number-iterations* 0)
	epilog::*continue-til-minimum*
	(epilog::*qa-start-real-time* 0)
	(epilog::*question-answer* 'epilog::no-question)
	epilog::*question*
	epilog::*normed-question*
	epilog::*solutions*
	epilog::*number-answers*
	epilog::*wh-pattern*
	epilog::*e-pattern*
	epilog::*matches*
	epilog::*e-matches*
	epilog::*qa-agenda*
	(epilog::*max-depth* 0)
	epilog::*max-wh-depth*
	epilog::*qa-only-try*
	epilog::*ok-to-rearrange*
	epilog::*current-attempt*
	epilog::*wff-subgoals*
	epilog::*proof*
	epilog::*disproof*
	lib::*question-ckpt*
	(lib::*num-anon* 0)		; no anon checkpoints in simulations.
	(lib::*max-anon* 0)
	lib::*oldest-anon*
	(*current-netname* '/)

	;; tweakables that get tweaked by the program itself,
	;; rather than by the user.  As for the above special
	;; variables, don't want simulation to change its
	;; parent's value.  Values are the defaults given in the
	;; tweakable declarations.
	;;   Note: if the user makes a tweak, the tweak won't
	;; get propagated to simulations.  That might be bad.
	(response::*filter-threshold* 100)
	(epilog::*qa-iterations* 40)
	(epilog::*question-effort* 3)

	;; (epilog::*print* nil) ; turn off printing in simulations
	)
     (eval form)))


