;       Last change    August 1993
(in-package epilib)

(defvar *symbol-number* 0)
(defvar *t-symbol-number* 0)

(defvar *temp-table*
  (initialize-element-table "Temporary table for formulas and subformulas")
  "Hash table of atoms")
(defvar *insertion* nil)

(tweakable '*input-array-expansion-size* 100
  "Amount to increase input array size by when full" '(integer 1 1000))
(traceable 'memory "Shows expansion of input array") 

(defstruct environment
  name            ;; Just a debugging aid.
  (input-index -1)
  (input-array (make-array
		(list epilib::*input-array-expansion-size*)  
		:initial-element nil
		:adjustable t))
  (main-net (make-hash-table :test #'equal))
  (topic-net (make-hash-table :test #'equal))
  (equivalence-table (initialize-element-table
  		      "Table of equivalence sets"))
  (element-table (initialize-element-table
		  "Hash table for formulas and subformulas"))
  use-environments ;; list of other environments whose knowledge this
                   ;; environment can use

  specialist-info ;; an alist for specialists to store their own
                  ;; special-purpose information in.
  )

(defvar *environment* (make-environment :name "Main Environment"))

(defun new-index (&aux old) 
  (increment 'environment-input-index *environment*)
  (when (>= (environment-input-index *environment*)
	  (setq old
	    (array-dimension (environment-input-array *environment*)
			     0)))
	(if (traced-p 'memory)
	    (print-line "Expanding input array by " *input-array-expansion-size*
			" elements"))
        (setf (environment-input-array *environment*)
	  (adjust-array (environment-input-array *environment*)
			(list (+ old *input-array-expansion-size*))
			:initial-element nil) ) )
  (environment-input-index *environment*))

(defun add-to-input-array (input-index name)
  (change-array (environment-input-array *environment*) input-index
		(append (geta (environment-input-array *environment*)
			      input-index)
			(list name))))


(defun new-symbol (&optional str temp &aux
			     num tmp)
  (if temp
      (setq num (change-var '*t-symbol-number* (1+ *t-symbol-number*)))
    (setq num (change-var '*symbol-number* (1+ *symbol-number*))))
  (setq tmp (build-symbol (append (if temp '(T-)) (list str num)) 'epilog))
  (setf (symbol-plist tmp) nil)
  (unless temp 
      (start-checkpoint 'remove-wff tmp))
  tmp)

(defun remove-wff (item)
  ; also remove from hash table?
  (setf (symbol-plist item) nil)
  )

; save-info creates a new symbol if necessary for item, and stores it
; in the hash table (or stores it under name if given). The name to refer
; to that item by is returned

(defun save-info (item name type &aux
		       (oldname (get-element item (environment-element-table *environment*)))
		       newname)
  (cond					; if both name, and oldname, equate them 
   ((and name oldname)
    (change-property name 'alternate-names
		     (push oldname (get name 'alternate-names))) ;; was-getp
					; already saved so no need to duplicate
    oldname)
   
					; if name, and no oldname, save item with name
   (name 
    (put-element item name (environment-element-table *environment*) nil)
    name)
   
					; no name, but oldname
   (oldname oldname)
   
   ((and (not *insertion*) (setq oldname (get-element item *temp-table*)))
    oldname)
   
					; if no name, and no oldname, build a new name and save it
   (t
    (setq newname (new-symbol type (if *insertion* nil t)))
    (setp newname 'entity-type type)
    (if *insertion*
	(put-element item newname (environment-element-table *environment*) nil)
      (put-element item newname *temp-table* t))
    newname)
   )
  )

(defun make-fn-arg (fn &rest args &aux new)
  (setq new (save-info (append (list fn) args) nil 'term))
  (setp new 'term-type 'function)
  (setp new 'operator fn)
  (setp new 'arguments args)
  new)


(defun temp-cleanup ()
  (setq *temp-table*
	(initialize-element-table 
	 "Temporary table for formulas and subformulas"))
  )

;;;; Reading and writing of permanent memory

(defvar *perm-memory-output* nil)
(defvar *perm-items* nil)
(defvar *perm-probs* nil)

; this all assumes that same hierarchy and indicators are loaded

; need to save hash table, both classification tables, all wffs, terms, probs,
; preds, ops, etc

(defun write-perm-memory (&optional (file "perm-memory"))
  (setq *perm-memory-output* (open file :direction :output
				   :if-exists :supersede
				   :if-does-not-exist :create))
  (format *perm-memory-output* "(in-package epilog)~%")
  (setq *perm-items* nil)
  (setq *perm-probs* nil)
  ; save specialist representations - each must use their own method
  (write-specialist-info)
  ; save symbol number
  (write-perm-symbol-number)
  ; go through all classifications, save them
  ; keep list of wffs
  (write-perm-classes 'main)
  (write-perm-classes 'topic)
  ;save normalization hash table
  (write-hash-table)
  ;; save hierarchy nodes, even ones that didn't occur in any wffs.
  ;; (otherwise we get inconsistent hierarchies when file is read back
  ;; in).
  ; then save each wff, keeping list of concepts
  (write-perm-items *perm-items*)
  ; save probabilities
  (write-perm-items *perm-probs*)
  (close *perm-memory-output*)
  )

(defun wpm (&rest args) (apply 'write-perm-memory args))

(defun write-perm-classes (flag &optional (subnet (if (equal flag 'main)
						      (environment-main-net *environment*) (environment-topic-net *environment*))) 
				prefix)
  (maphash #'(lambda (key entry)
		     (when entry
		     (if (hash-table-p entry)
			 (write-perm-classes flag entry (list key))
			 (write-perm-class flag (if prefix
					       (append prefix (list key))
					       key)
					   entry)
		     )))
	   subnet)
  )

(defun add-probs (l &aux pl)
  (when (listp l)
	(setq pl (remove-if #'(lambda (a) 
			    (or (numberp a) (getp a 'wff-type))) l))
	(setq *perm-probs* (union *perm-probs* pl))
	(mapcar #'(lambda (a) (add-probs (getp a 'prob))) pl))
  )

(defun write-perm-class (flag key entry)
  (write-perm-line "epilib::read-perm-class" (list flag key entry))
  ; save new wffs in entry in *perm-wffs*
  (dolist (e entry) (add-probs (getp e 'prob)))
  )

(defun write-hash-table (&aux (table (element-table-vector (environment-element-table *environment*)))
			      entry)
  (write-perm-line "epilib::read-hash-table" 
		   (list (element-table-load (environment-element-table *environment*))
			 (element-table-size (environment-element-table *environment*))
			 (element-table-size-index (environment-element-table *environment*))
			 (element-table-rehash-threshold (environment-element-table *environment*))
			 (element-table-current-rehash (environment-element-table *environment*))
			 ))
  (do ((i 0 (+ i 1)))
      ((>= i (element-table-size (environment-element-table *environment*))))
      (setq entry (aref table i))
      (when entry
	    (write-perm-line "epilib::read-hash-table-entry" (list i entry))
	    ; add elements to permanent list?
	    (dolist (e entry) (add-entry (cdr e)))
	    ))
  )

(defun add-entry (w &aux p)
  (setq *perm-items* (append *perm-items* (list w))) ; put wffs at end
		    
  (case (entity-type w)
	(wff (if (getp w 'variable) (pushnew (getp w 'variable) *perm-items*))
	     (if (getp w 'quantifier) 
		 (pushnew (getp w 'quantifier) *perm-items*))
	     (dolist (a (getp w 'arguments)) (pushnew a *perm-items*))
	     (if (and (setq p (getp w 'pred)) (symbolp p))
		 (pushnew p *perm-items*))
	     (if (getp w 'operator) (pushnew (getp w 'pred) *perm-items*)) )
	(pred (if (getp w 'operator) (pushnew (getp w 'operator) *perm-items*))
	      (dolist (a (getp w 'arguments)) (pushnew a *perm-items*)) )
	(term (if (getp w 'operator) (pushnew (getp w 'operator) *perm-items*))
	      (dolist (a (getp w 'arguments)) (pushnew a *perm-items*))  )
	(op (if (getp w 'operator) (pushnew (getp w 'operator) *perm-items*))
	    (dolist (a (getp w 'arguments)) (pushnew a *perm-items*))  )
	)
  )

(defun write-perm-items (items) 
  (dolist (w items) 
	  (when (and (symbolp w) (not (numberp w)) (not (stringp w))
		     (not (member w '(and or implies ~> <=>))))
		(write-perm-item w)
		; handle skolemized variables as well
		(if (and (getp w 'var) (not (member (getp w 'var) items)))
		    (write-perm-item (getp w 'var)))
		)))
; need to test for strings/numbers/etc here?

(defun remove-special-forms (plist &aux nlist newlist)
  (setq nlist plist) 
  (do ((i (car nlist) (car nlist)) 
       (j (cadr nlist) (cadr nlist))) 
      ((and (null i) (null j)))
	  (setq nlist (cddr nlist))              
	  (unless (or (functionp j) (arrayp j))
		  (setq newlist (append newlist (list i)))
		  (setq newlist (append newlist (list j)))
	      )
      )
  newlist)

(defun write-perm-item (item)
  ; eliminate the system special stuff here - it screws up when reread
  (write-perm-line "epilib::read-perm-item" 
		   (list item #-lucid (remove-special-forms (symbol-plist item))
			 #+lucid (symbol-plist item)))
  )

(defun write-perm-symbol-number ()
  (write-perm-line "(lambda (x) (setq epilib::*symbol-number* x))" (list *symbol-number*)))

(defun write-specialist-info ()
  (dolist (s *active-specialists*)
	  (write-perm-line "use-spec" (list s))
	  ; there is a permanent-memory routine on spec property list, use
	  )
  )

(defun write-perm-line (rtn args)
  ; write to perm file
  (format *perm-memory-output* "(~A ~{'~S ~})~%" rtn args)
  )


; Reading it back in and setting things up

(defun read-perm-memory (&optional (file 'perm-memory))
  ; restore symbol number
  ; restore each concept
  ; restore each wff
  ; restore classifications
  ; just load the file - all the commands should be there
  ; NOTE - this cannot be done after other wffs have been entered!!!
  (if (> *symbol-number* 0)
      (print-line "Cannot load permanent memory after a wff has been added")
      (load file)
      )
  )

(defun rpm (&rest args) (apply 'read-perm-memory args))

(tweakable '*memory-load-specs* t
  "Indicates whether to call specialists to reload formulas when reading in memory file"
  'boolean)

(defun read-perm-item (item plist)
  (setf (symbol-plist item) (union (symbol-plist item) plist :test 'equal))
  (when (and *spec-enter* *memory-load-specs*
	     (equal (getp item 'entity-type) 'wff)
	     (getp item 'c-prob)
	     (= (getp item 'c-prob) 1)
	     )
	(spec-enter item))
  )

(defun read-perm-class (key class wffs)
  ; shortcut class process to save all wffs at once
  (shortcut-store-class (if (equal key 'main) (environment-main-net
					       *environment*)
			  (environment-topic-net *environment*))
			class wffs)
  )

(defun shortcut-store-class (subnet class wff-list &aux l class+)
  (when (listp (car class)) 
        (setq class+ (cdr class))
        (setq class (car class))
        )
  ; get current contents of that classification - will be either a list
  ; of propositions or another hash table
  (setq l (get-class-from-net subnet class))

  ; if it is nil, and there is more classifying to do, build a subnet hash
  ; table and store it
  (when (and (null l) class+)
      (setq l (make-hash-table :test #'equal))
      (change-hash class subnet l)
      )  

  (if (listp (car class+)) 
      (setq class+ (car class+)))

  ; if there is a further level, continue classifying with new subnet,
  ; else just add this item to the current list of propositions
  (if (hash-table-p l)
      (shortcut-store-class l class+ wff-list)
      (change-hash class subnet wff-list))
  )

(defun read-hash-table (load size size-index rehash-threshold current-rehash)
  (change-field 'element-table-load (environment-element-table *environment*) load)
  (setf (element-table-size (environment-element-table *environment*)) size)
  (setf (element-table-size-index (environment-element-table *environment*)) size-index)
  (setf (element-table-rehash-threshold (environment-element-table *environment*)) rehash-threshold)
  (setf (element-table-current-rehash (environment-element-table *environment*)) current-rehash)
  (setf (element-table-vector (environment-element-table *environment*)) 
	(make-array size :initial-element nil))
  )

(defun read-hash-table-entry (index entry)
  (change-array (element-table-vector (environment-element-table *environment*)) index entry)
  )

(defun save-equivalence (item subnet-info equivalence-set)
  ; see if there is already a set there - if so, replace
  (if (get-equivalence item subnet-info)
      (delete-element (list item subnet-info) (environment-equivalence-table *environment*)))
  (put-element (list item subnet-info) equivalence-set (environment-equivalence-table *environment*))
  )

;; ANK this does look in the use-environments, but fails to come up
;; with equivalences that could be discovered by transitively
;; combining information from multiple environments.
(defun get-equivalence (item subnet-info)
  (remove-duplicates
   (apply #'append
	  (map-environments #'(lambda (env)
				(get-element (list item subnet-info)
					     (environment-equivalence-table env)))
			    *environment*))))
