;       Last change    August 1993
(in-package epilib)
; Retrieval routines

; The classification strategy is described in class.lisp. These
; rotuines use that strategy to retrieve desired information.

; Useful routines:
;    get-class class
;      returns either a list of propositions or a subnet hash-table
;      class may be a list of pairs, in which case, get-class will
;      recursively call itself with the hashtables it obtains to get the
;      final list of propositions.
;    retrieve concept topic super sub print-subnets subnet
;      retrieves the propositions in subnet for the topic under the 
;      constant concept. If print-subnets is t, subnets will be 
;      exhaustively printed - otherwise they will only be noted.
;      If sub is specified, topics beneath topic in the hierarchy are
;      also printed (also concepts beneath concept in the type hierarchy); 
;      if super, those above as well
;      If no topic is specified, all topics are printed
;    ret concept topic super sub print-subnets subnet
;      just a shorter name for retrieve
;    display-concept concept topic super sub print-subnets subnet
;      just like retrieve except that the propositions are printed with
;      topic headings instead of collected and returned in a list
;    display concept topic super sub print-subnets subnet
;      just a shorter name for display-concept
;    get-everything (optional subnet print-subnets)
;      returns a list of all the propositions in the subnet, including
;      subnets if print-subnets is t
;    show-everything (subnet indent print-subnets)
;      prints everything in the subnet, showing their classifications
;      if print-subnets is t, subnets beneath this are printed exhaustively;
;      otherwise they are only noted


;; apply function fn to the environment env and all of its subordinate
;; environments (the contents of its environment-use-environments
;; field and, recursively, their subordinate environments).  Return a
;; list of results, one for each environment.

(defun map-environments (fn env)
  (cons (funcall fn env)
	(apply #'append
	       (mapcar #'(lambda (env)
			   (map-environments fn env))
		       (environment-use-environments env)))))

;; get-class gets the current contents of a concept-topic pair (or
;; list of pairs for subnet classifications) in all of the KBs used by
;; an environment (*environment* by default).  Uses the environment's
;; main-net unless the topic-net argument is t, in which case it uses
;; the topic-net.

(defun get-class (class
		  &optional (topic-net nil) (env *environment*))
  (funcall #'merge-class-lists
	 (map-environments (if topic-net
			       #'(lambda (env)
				   (get-class-from-net (environment-topic-net env)
						       class t))
			     #'(lambda (env)
				 (get-class-from-net (environment-main-net env)
						     class t)))
			   env)))

;; The list of formulas in a classification is ordered by complexity
;; and probability.  lists is a list of lists, each list having
;; retrieved from an environment by get-class-from-net.  This function
;; combines those lists into one sorted list.
(defun merge-class-lists (lists)
  (let ((cars (remove-if #'null
			 (mapcar #'car lists)))
	(cdrs (remove-if #'null
			 (mapcar #'cdr lists))))
    (when cars
      (append (sort cars #'proposition-order-lessp)
	      (merge-class-lists cdrs)))))

; get-class-from-net gets the current contents of a concept-topic pair
; in subnet if the class is a list of concept-topic pairs, these will
; be followed into the subnets if posible

(defun get-class-from-net (subnet class &optional (get-all-equiv t)
			  &aux res class+ l)
  (when (listp (car class))
        (setq class+ (cdr class))
	(setq class (car class))
	)

  (dolist (c (if (and (symbolp (car class)) get-all-equiv)
		 (list (car class))
		 (list (car class))))
	  ; get current contents for class given (first level) for this subnet
	  (setq l (geth (append (list c) (cdr class)) subnet))

	  ; if this entry indicates a hash table, and there is more 
	  ; classification info given, continue following it. Else just 
	  ; hand back th entry.
	  (if (and (hash-table-p l) class+)
	      (setq res (append res (get-class-from-net l class+)))
	      (if (hash-table-p l)
		  (setq res (append res (list l)))
		  (setq res (append res l)))
	      )
	  )
  (remove-duplicates res)
  )

; print-prop list prints each proposition in a list on a separate line, 
; preceeded by indent # of spaces

(defun print-prop-list (indent plist &optional (probs t))
  (dolist (l plist)
	  (indent indent)
	  (print-it l ": " (getp l 'PRINT))
	  (when probs
		(print-it "     Probability: " (getp l 'C-PROB)))
	  (if (getp l 'PROOF) (print-it "     (proof)"))
	  (if (getp l 'DISPROOF) (print-it "     (disproof)"))
	  (if (getp l 'assumption) (print-it "    (assumption)"))
	  )
  )

(defvar said nil "keeps track of topics in listing so not repeated")

; print-topic gethers and prints all topically organized information
; for concpet for topic, and all topics under topic in the topic hierarchy
; subnets are not printed for now, but are tested for

(defun print-topic (indent concept topic path subnets sub
			   &aux plist)
  (setq plist (get-class (list concept topic) t))
  (when plist
	; print the topics leading up to this item
	; to them, and then print the propositions
	; check leading topics to see if they have already been printed
	; (only print topics that actually have something under them)
	(dolist (p path)
		(unless (member (cdr p) said)
			(indent (car p))
                        (print-it (cdr p)) 
			(push (cdr p) said)))
	(indent indent) (print-it topic) 
	(push topic said)

	; if subnet here,  print contents
	(if (hash-table-p (car plist))
	    (if subnets
		(mapcar #'(lambda (a) (show-everything a (+ indent 2) t))
			plist))

	  ;; otherwise a list of propositions - 
	  (print-prop-list (+ indent 2) plist))
	)

  (when sub
  ; repeat for any topic node under topic in the hierarchy
  (dolist (tp (get-all-sub topic)) 
	  (print-topic (+ indent 4) concept tp 
		       (append path (list (cons indent topic))) subnets sub)
	  ))
  )

; get-topic gets the propositions under concept and topic, from the
; topic-net of all accessible environments.
; if sub is specified, subtopics are alos obtained; if subnets is 
; specified, the contents of subnets is also obtained

(defun get-topic (concept topic subnets sub &aux plist rlist)
  (setq plist (remove-duplicates (apply 'append 
		(mapcar #'(lambda (a) (get-class (list a topic) t))
			(append (list concept) (equivalent-items concept))))))
  (if (hash-table-p (car plist))
      (if subnets
	  (setq rlist (apply 'append (mapcar #'(lambda (a) (get-everything a))
					     plist))))
      (setq rlist plist))

  ; get sub topics if requested
  (when sub
  ; repeat for any topic node under topic in the hierarchy
  (dolist (tp (get-all-sub topic)) 
	  (setq rlist (append (get-topic concept tp subnets sub)
			      rlist))
	  ))
  rlist)

; retrieve gets requested information stored about concept from the
; topic net.  If topic is specified, it is used, otherwise all topics
; are.  subnets is t if the contents of subnets should be included,
; nil otherwise sub indicates that subtopics should also be included,
; sup that super topics should

(defun retrieve (concept
		 &optional (topic nil) (super nil) (sub nil) (subnets nil)
		 &aux rlist skol
		      (usec (if (and (symbolp concept) 
				     (setq skol (getp concept 'skol-const)))
				skol
			      concept)))

  (if topic
      ;; if a topic is specified, use it
      (let ()
	;; get super topics if requested
	(if super
	    (let ((suplist (get-all-super topic)))
	      (do ((suptop (car suplist) (car suplist)))
		  ((null suptop))
		(setq suplist (cdr suplist))
		(setq suplist (union suplist (get-all-super suptop)))
		(setq rlist (append 
			     (get-topic usec suptop subnets nil)
			     rlist))))
	  )
	;; get this topic and subs if requested
	(setq rlist (append (get-topic usec topic subnets sub) rlist))
	)

    ;; get all topics
    (dolist (tp (get-all-sub 'TP.TOPICS)) 
      (setq rlist (append (get-topic usec tp subnets t)
			  rlist)))
    )
  ;; get for sub concepts if requested
  (when sub
    (dolist (item (get-all-sub usec))
      (setq rlist (append
		   (retrieve item topic nil sub subnets)
		   rlist))))
  ;; get for super concepts if requested
  (when super
    (dolist (item (get-all-super usec))
      (setq rlist (append
		   (retrieve item topic super nil subnets)
		   rlist))))
  (remove-duplicates rlist)
  )

; ret - short form of retrieve
(setf (symbol-function 'ret) #'retrieve)

; display concept prints all information about a concept for a given
; topic if specified, or for all topics.
; This is exactly like retrieve, except that the propositions are
; printed instead of returned in a list.

(defun display-concept (concept &optional (topic nil) (super t) (sub t)
			 (subnets nil) equ usec f skol)
			 
  (setq usec (if (and (symbolp concept) 
		      (setq skol (getp concept 'skol-const)))
		 skol
		 concept))
  (if (equal topic 'none) (setq topic nil))
  (if (equal super 'none) (setq super nil))
  (if (equal sub 'none) (setq sub nil))
  (if (equal subnets 'none) (setq subnets nil))
  (setq said nil)
  (print-line concept)
  (unless (equal usec concept) (print-it " (" usec ")"))
  (when (and (setq f (find-sort usec)) 
	     (not (member f '(pred wff term))))
	(print-it "  Sort: " f))
  (when (setq equ (remove concept (equivalent-items concept)))
	(print-it "  Equivalent to: " (mapcar 'print-info equ)))
  (if topic
      (let ()
      (if nil ;super
	  (let ((suplist (get-all-super topic)))
	  (do ((suptop (car suplist) (car suplist)))
	      ((null suptop))
	      (setq suplist (cdr suplist))
	      (setq suplist (union suplist (get-all-super suptop)))
	      (print-topic 4 usec suptop nil subnets nil))))
      (print-topic 4 usec topic nil subnets t)
      )
      (dolist (tp (get-all-sub 'TP.TOPICS)) 
	      (print-topic 4 usec tp nil subnets t))
      )
  (when sub
	(dolist (item (get-all-sub usec)) 
		(display-concept item topic nil sub subnets)))
  (when super
	(dolist (item (get-all-super usec))
	     (display-concept item topic super nil subnets)))
  )

(set-display-function 'node 'display-node "Displays wffs about a constant")
(setq *display-default* 'node)

(defun display-node (brief &rest args &aux
			   (concept (car args))
			   (topic (cadr args))
			   (super (not brief))
			   (sub (not brief))
			   (subnets (not brief)) )
  (if (getp concept 'wff-type)
      (pr-formula brief concept)
      (display-concept concept topic super sub subnets))
  )

; get-everything gets all propositions in a subnet. If subnets is t,
; it recursively gets all propositions in subnets as well.

(defun get-everything (subnet &optional (subnets t)
		       &aux rlist)
  (maphash #'(lambda (key entry)
	             (declare (ignore key))
		     (when entry
		     (if (hash-table-p entry)
			 (if subnets
			     (setq rlist (append 
					  (get-everything entry subnets)
					  rlist)))
			 (setq rlist (append entry rlist)))
		     ))
	   subnet)
  (remove-duplicates rlist)
  )

;; show-everything shows all things organized in the knowledge base, 
;; including the contents of subnets (if subnets is t)

;; Corrected discrepancy with the doc: put all the arguments as
;; optionals with the default assignment of subnet as required by the
;; documentation. FM
(defun show-everything (&optional (subnet (environment-main-net *environment*)) (indent 0) (subnets nil))
  (maphash #'(lambda (key entry)
	       (when entry
		 (indent indent)
		 (print-it key) 
		 (if (hash-table-p entry)
		     (if subnets
			 (show-everything entry (+ indent 4) subnets)
		       (let () 
			 (indent indent)
			 (print-it "    subnet")))
		   (print-prop-list (+ indent 4) entry)
		   )))
	   subnet)
  )

(defun get-em-all (&optional brief &rest args)
  (declare (ignore args))
  (print-line "Accessible environments: ")
  (map-environments #'(lambda (env)
			(print-it (environment-name env))
			(print-it ", "))
		    *environment*)
  (print-line)
  (print-line "Contents of current environment "
	      (environment-name *environment*) ":")
  (show-everything (environment-main-net *environment*)
		   0
		   (not brief)))

(defun get-em-topic (&optional brief &rest args)
  (declare (ignore args))
  (print-line "Accessible environments: ")
  (map-environments #'(lambda (env)
			(print-it (environment-name env))
			(print-it ", "))
		    *environment*)
  (print-line)
  (print-line "Contents of current environment "
	      (environment-name *environment*) ":")
  (show-everything (environment-topic-net *environment*)
		   0
		   (not brief)))

(set-display-function 'wffs 'get-em-all
  "Displays ALL wffs in the system, using the main classification")
(set-display-function 'wffs-by-topic 'get-em-topic
  "Displays all wffs (except schemas) in the system, using the topic classification")

(defun instance-type (garg &aux
		 (arg (cond ((getp garg 'skol-const)) (t garg))) )
  (grab-types arg))

(defun description (garg &aux dlist wargs
			 (arg (cond ((getp garg 'skol-const)) (t garg))) )
  (dolist (w (get-class (list arg 'episode-specialist)))
	  (setq wargs (getp w 'arguments))
	  (when (and (not (neg-p w)) (equal (last1 wargs) arg))
	        (pushnew (getp (car wargs) 'print) dlist :test #'equal))
	  )
  dlist)

(defun content (garg &aux dlist wargs
			 (arg (cond ((getp garg 'skol-const)) (t garg))) )
  (dolist (w (get-class (list arg '_ 'object)))
	  (setq wargs (getp w 'arguments))
	  (when (and (not (neg-p w)) (equal (last1 wargs) arg))
	        (pushnew (getp (car wargs) 'print) dlist :test #'equal))
	  )
  dlist)

(defun show-input-array (&optional brief
			 &rest junk
			 &aux wff-name wffs
			      (bound (- (array-dimension
					 (environment-input-array *environment*)
					 0)
					1)))
  ;;
  ;; Displays narrative facts as stored in the Input Array.
  ;;
  (declare (ignore brief junk))
  (print-line " == Input Array ==" *nl* )
  (do ((index 0 (1+ index)))
      ((> index bound) (return))
      (setq wffs (geta (environment-input-array *environment*) index))
      (if (null wffs) (return))
      (setq wff-name (car wffs))
      (print-line "   [" index "] " wff-name ": " (getp wff-name 'PRINT))
      (dolist (w (cdr wffs)) (print-line "       " w ": " (getp w 'PRINT))) 
      )
 ) ; end SHOW-INPUT-ARRAY

(set-display-function 'input-array 'show-input-array 
  "Displays narrative facts as stored in the input array")
(set-display-function 'input 'show-input-array 
  "Displays narrative facts as stored in the input array")


(defun print-wff-type (brief wff-name)
  (declare (ignore brief))
  (cond ((not (getp wff-name 'print))
	 (print-line "%% Error: " wff-name " does not exist **" ))
	((getp wff-name 'parents) (print-line "** Inferred formula **"))
	((getp wff-name 'input-index) (print-line "** Given by the story **" ))
	((getp wff-name '-keys)
	 (cond ((getp wff-name 'meaning-postulate)
		(print-line "** Meaning postulate **"))
	       ((getp wff-name 'simplification-schema)
		(print-line "** Simplification schema **"))
	       (t (print-line "** Given as world knowledge **") )))
	(t (print-line "** This is only part of a wff **"))
	)
  NIL)

(defun pr-formula (brief &rest wff-names) 
  (dolist (wff-name wff-names) (formula brief wff-name)))

(set-display-function 'wff 'pr-formula "Prints actual formula")
(set-display-function 'formula 'pr-formula "Prints actual formula")

(defun formula (brief wff-name &aux p)
  ;; 
  ;; Given a wff-name, displays its formula.
  ;;
  (IF (setq p (GETP WFF-NAME 'PRINT))
      (print-line " " WFF-NAME ": " p)
      (print-line " %% Error: " wff-name " does not exist " ))
  (print-wff-type brief wff-name)
  (when (not brief)
	(prob wff-name)
	(when (getp wff-name 'controlled-vars)
	      (print-line "  Controlled variables: " 
			  (getp wff-name 'controlled-vars)))
	(when (getp wff-name '-trigger)
	      (print-line "  Negative trigger keys:")
	      (dolist (k (getp wff-name '-trigger))
		      (print-line "    " k ": " (getp k 'print)))
	      (print-line "  Positive trigger keys:")
	      (dolist (k (getp wff-name '+trigger))
		      (print-line "    " k ": " (getp k 'print)))
	      ))
  NIL)


(defun pr-prob (&optional brief &rest wff-names) 
  (declare (ignore brief))
  (dolist (wff-name wff-names) (prob wff-name)))

(set-display-function 'prob 'pr-prob "Prints lower subjective probability")

(defun prob (wff-name)
  ;; 
  ;; Given a wff-name, displays its lower subjective probability.
  ;;
  (if (getp wff-name 'PRINT)
     (print-line " " wff-name "'s lower subj. prob. = " (getp wff-name 'PROB) 
      " which calculates to " (getp wff-name 'C-PROB) )
     (print-line " %% Error: " wff-name " does not exist " ))
 )



