;       Last change    August 1993
(in-package epilib)

;;Substitution routines - these routines do substitutions in a proposition
;; and return a new one.

(defvar *multiple-subs* nil)
(defvar *subst-prob* 1)

; Some routines to make it easier to work with propositions

; make-prop creates a new proposition with the appropriatee bracketing 
;; structure given certain key parts

(defun make-prop (negate &key (pred nil) (args nil) (quantifier nil) 
			 (variable nil) (restriction nil) (main-clause nil) 
			 &aux p)
  (setq p
  (if quantifier
      (if restriction
	  (list quantifier variable restriction main-clause)
	  (list quantifier variable main-clause))
      (append (list (car args) pred) (cdr args))
      ))
  (if negate
      (cons 'not  p)
      p)
  )

(defun subs-item (item subs)
  (cond ((symbolp item)
	 (case (entity-type item)
	       (wff (subs-wff item subs))
	       (pred (subs-pred item subs))
	       (term (subs-term item subs))
	       (op (subs-op item subs))
	       (otherwise (subs-term item subs)))
	 )
	(t item))
  )

;;; Subs-prop performs a set of substitutions on a proposition.
;;; It is non-destructive on the proposition, creating a new copy in 
;;; all cases.
;;;
;;;	prop is a normalized proposition (or literal)
;;;	subs is the list of substitutions to perform.  
;;;
;;;	Takes a normalized wff and returns a non-normalized wff
;;;
(defun subs-wff (prop subs &aux sv)

  (if (not (equal (setq sv (subs-variable prop subs)) prop))
      (if (symbolp sv)
	  (subs-wff sv subs)
	  sv)
  (case (getp prop 'wff-type)

	((prefix causal episodic logical)
	 (let ((args (getp prop 'arguments))
	       (pred (getp prop 'pred))
	       newp newa)
	      (setq newp (subs-pred pred subs))
	      (setq newa (mapcar #'(lambda (a) 
					   (subs-item a subs))
				 args))
	      (append (if (neg-p prop) (list 'not))
		      (list (car newa) newp)
		      (cdr newa))
	      ))

	(quasi-quoted-expression
	 (list 'qquote (subs-wff (getp prop 'quote-wff) subs)))

	(quoted-expression (getp prop 'print))

	(quantified
	 ; if the variable is one of the things being sub'd for,
	 ; leave the quantifier and variable out of the result
	 (let* ((quant (getp prop 'quantifier))
		(var (getp prop 'variable))
		(rest (getp prop 'restrict))
		(main (getp prop 'main))
		sub newvar)

	       (setq sub (assoc var subs))
	       (setq newvar (subs-term (subs-variable var subs) subs))
	       (if (and sub 
			(or (not (epi-variable newvar))
			    (not (equal (caddr sub) t))
				))
		   (let ()
			(setq *subst-prob* (combine-probs *subst-prob*
					  (cond ((getp quant 'prob)) (t 1))))
		   (if rest ; problem - how to combine without quantifier?
		       ; use +ve -ve context to help decide this?
		       (combine-subgoal (if (equal quant 'A) 'implies 'and)
					(subs-wff rest subs)
					(subs-wff main subs))
		       (subs-wff main subs)) )
			
		   ; otherwise, keep quantifier, continue subing
		   (append (list quant newvar)
			   (if rest (list (subs-wff rest subs)))
			   (list (subs-wff main subs)))
		   )
	       )
	 )

	(modified (subs-modified prop subs))
	((variable constant)
	 (if (and (neg-p prop) (getp prop 'print))
	     (let ((res (subs-wff (cadr (getp prop 'print)) subs)))
		  (if (listp res)
		      (append (list 'not) res)
		      (list 'not res)))
	     prop))
	(otherwise prop)
	))
  )


;;; Subs-term does substitutions for a term.
;;;
;;;	Takes a normalized term and returns a non-normalized term.
;;;
(defun subs-term  (term subs &aux sv)
  (if (not (equal (setq sv (subs-variable term subs)) term))
      (let () 
      (if (and (symbolp sv) nil 
	       )
	  (subs-term sv subs)
	  sv) )
  (cond ((symbolp term)
	 (case (getp term 'term-type)
	       (constant term)
	       (quoted-expression term) 
	       ((record quasi-quoted-expression)
		(subs-quoted-item (getp term 'print) subs))
	       (wff (subs-wff term subs))
	       ((function modified) (subs-modified term subs))
	       (otherwise term)
	       ))
	(t term))
      )
  )

(defun subs-quoted-item (qlist subs)
  (cond ((listp qlist) (mapcar #'(lambda (a) (subs-quoted-item a subs))  qlist))
	(t (subs-item qlist subs)))
  )


(defun subs-pred (pred subs &aux sv)
  (if (not (equal (setq sv (subs-variable pred subs)) pred))
      (if (symbolp sv)
	  (subs-pred sv subs)
	  sv)
  (cond ((symbolp pred)
	 (case (getp pred 'pred-type)
	       (constant pred)
	       (lambda
	        ; 2 kinds - lambda op, and simple lambda
	        (if (getp pred 'operator)
	            (subs-modified pred subs)
	            (subs-wff pred subs))
	        )
	       (modified (subs-modified pred subs))
	       (otherwise pred)
	       ))
	((listp pred) 
	 (append (list (car pred)) (mapcar #'(lambda (a) (subs-variable a subs)) (cdr pred))))
	(t pred))
      )
  )

(defun subs-op (op subs &aux sv)
  (if (not (equal (setq sv (subs-variable op subs)) op))
      (if (symbolp sv)
	  (subs-op sv subs)
	  sv)
  (cond ((symbolp op)
	 (case (getp op 'op-type)
	       (constant op)
	       (modified (subs-modified op subs))
	       (otherwise op)
	       ))
	(t op))
      )
  )

(defun subs-variable (var subs &aux sub)
  (if (setq sub (assoc var subs :test 'equal))
      (cadr sub)
      var)
  )

(defun subs-modified (item subs)
  (append (list (subs-op (getp item 'operator) subs))
	  (mapcar #'(lambda (a) (subs-item a subs)) (getp item 'arguments)))
  )

;;; 'External' routines for the two above--they include some basic
;;; argument checks.
;;;
(defun substitute-prop (prop subs &optional subs-2 (subs-num 1) (mult-subs t)
			     &aux result)
  (setq *subst-prob* 1)
  (if mult-subs (setq *multiple-subs* t))
  (setq result (subs-wff prop (if (equal subs-num 1) subs subs-2)))
  (setq *multiple-subs* nil)
  result
)

(defun substitute-term (term subs &optional subs-2 (subs-num 1) (mult-subs t)
			     &aux result)
  (if mult-subs (setq *multiple-subs* t))
  (setq result (subs-term term (if (equal subs-num 1) subs subs-2)))
  (setq *multiple-subs* nil)
  result
)

; make-sub-element makes a substitution element consisting of the new argument
; and the old. If there is no change, nil is returned.

(defun make-sub-element (newarg oldarg &optional which)
  (if (equal newarg oldarg)
      nil
      (if (null which)
          (list oldarg newarg)
          (list oldarg newarg which))
      )
  )    
 
; arg-sub-list takes two argument lists and returns the substitution list
; required to make them the same.

(defun arg-sub-list (newargs oldargs &optional which)
  (remove-if #'null (mapcar #'(lambda (a b)
                             (unless (null a) (make-sub-element a b which)))
                            newargs oldargs))
  )
 

