;;; The EpiShell commands package.  A miscellaneous collection of shell
;;; commands.

(in-package user)
(eval-when (compile load eval)   
(provide 'esh-cmds)
(require 'esh-io "epilog:eshell;io")
(defpackage esh-cmds 
  (:use lisp #+lucid lucid-common-lisp #-lucid common-lisp #+excl excl
	esh-io esh-interp lib)
  (:export 

          esh-apropos   ;lookup commands
	  cat	        ;concatenate files
	  echo          ;echo arguments
          os-shell	;start up an operating system shell
          os-csh	;start up an operating system csh
          drib          ;dribble 
          change-dir    ;change directory
          paf lpq ls    ;misc shell commands
          vi emacs less
	 ))
(unless (find-package 'esh-cmds) (make-package 'esh-cmds)))
(in-package esh-cmds)

(defun esh-apropos (&rest args &aux (cmd-specs (reverse *esh-command-list*)))
    "Apropos lookup of shell commands"
    (if args
        (let ((test (princ-to-string (car args))))
             (mapc #'(lambda (name-syns &aux (matched nil))
                         (do ((syn name-syns (cdr syn)))
	                      ((or (not syn) matched))
	                    (setq matched 
                                  (apropos-match test (car syn) name-syns)))
                         (if (and (not matched)
                                  (documentation (car name-syns) 'function))
                            (apropos-match test
                                (documentation (car name-syns) 'function)
                                 name-syns)))
                   cmd-specs))
       (print-msg "You should supply a string to match on."))
    nil 
)

(defun apropos-match (test syn name-syns 
                      &aux (matched (search test (prin1-to-string syn) 
                                            :test #'char-equal)))
    (when matched
         (prin1 (car name-syns))
	 (mapc #'(lambda (syn)
                     (princ " ")
                     (princ syn))
	       (cdr name-syns))
         (format *standard-output* "~35T")
         (print-msg (if (documentation (car name-syns) 'function)
                        (documentation (car name-syns) 'function))))
    matched
)

(defun cat (&rest args) "Concatentate files"
    (if args 
	(dolist (filename args)
		(if (setq filename (check-filename filename nil))
		    (with-open-file (file-stream filename 
						:if-does-not-exist nil)
                        (cond
                            (file-stream
   				 (do ((line (read-line file-stream nil esh-eof)
	      			          (read-line file-stream nil esh-eof)))
       				     ((eq line esh-eof))
				     (write-line line)))
                            (t (file-open-msg *standard-input* filename))))
		 ))
        (do ((line (read-line *standard-input* nil esh-eof)
	           (read-line *standard-input* nil esh-eof)))
            ((eq line esh-eof))
	    (write-line line)))
    nil
)


(defun drib (&rest arg) "Dribble screen i/o to a file"
   (if arg 
       (if (setq arg (check-filename (car arg) nil))
	   (dribble arg))
       (dribble))
   nil
)

(defun concat-args (args)
    (if args
        (subseq (reduce #'(lambda (str arg)
                   (concatenate 'string str " "
                           (cond ((not arg) "")
                                 ((stringp arg) arg)
                                 (t (write-to-string arg :pretty nil 
                                            :escape nil :case :downcase)))))
                  args :initial-value "") 1)
         "")
)

(defun echo (&rest args) "Echo arguments"
    (fresh-line)
    (princ (concat-args args))
    (fresh-line)
    nil
)


(defun os-cmd (cmd &rest args)
    #+lucid (lucid::run-unix-program cmd
	:arguments (mapcar #'(lambda (arg) 
                                   (if (not (stringp arg))
				       (string-downcase (princ-to-string arg))
                                       arg)) 
			   args)

	:input (if (get-filename *standard-input*) *standard-input*)
	:output (if (get-filename *standard-output*) *standard-output*)
	:error-output (if (get-filename *error-output*) *error-output*))
    #+allegro (excl:run-shell-command (squish (append (list cmd)
				 (mapcar #'(lambda (arg) 
                                   (squish (list " " (if (not (stringp arg))
				       (string-downcase (princ-to-string arg))
                                       arg)) )) args)))

	:input (if (get-filename *standard-input*) *standard-input*)
	:output (if (get-filename *standard-output*) *standard-output*)
	:error-output (if (get-filename *error-output*) *error-output*))
)

(defun os-shell (&rest args) "Start up a Unix shell (sh)"
    (if args
       (apply #'os-cmd `( "/bin/sh" "-c" ,(concat-args args)))
       (os-cmd "/bin/sh"))
)

(defun os-csh (&rest args) "Start up a Unix shell (csh)"
    (if args
       (apply #'os-cmd `( "/bin/csh" "-c" ,(concat-args args)))
       (os-cmd "/bin/csh"))
)

                                     
;;; In all of the following, use the shell to get
;;; file expansion and search paths.

(defun emacs (&rest args) "alias for: ! xemacs -font fixed-medium ..."
    (apply #'os-csh `( "xemacs" "-font" 
              "-misc-fixed-medium-r-normal--13-100-100-100-c-70-iso8859-1"
               ,@args))
)

(defun less (&rest args) "alias for: ! less -dm ..."
    (apply #'os-shell `( "less" "-dm" ,@args))
)

(defun paf (&rest args) "alias for: ! paf -format P ..."
    (apply #'os-shell `( "paf" "-format" "P" ,@args))
)


(defun lpq (&rest args)	"alias for: ! lpq ..."
    (apply #'os-shell `( "lpq" ,@args))
)


(defun ls (&rest args)	"alias for: ! ls ..."
    (apply #'os-shell `( "ls" ,@args))
)

(defun pwd (&rest args) "alias for: ! pwd ..."
    (apply #'os-shell `( "pwd" ,@args))
)

(defun vi (&rest args) "alias for: ! vi ..."
    (apply #'os-shell `( "vi" ,@args))
)

(defun change-dir (&rest args) "change directory"
       (let ((cdfn #+allegro #'excl:chdir
		   #-allegro #'user::cd
		   ))
	 (princ (if args (funcall cdfn (concat-args args))
		  (funcall cdfn)))
	 nil
))

