;;; The EpiShell read package
(in-package user)
(eval-when (compile load eval)   
(provide 'esh-read)
(require 'esh-io "epilog:eshell;io")
(defpackage esh-read 
  (:use lisp #+lucid lucid-common-lisp #-lucid common-lisp #+excl excl
	esh-io )
  (:export
	  *esh-take-char* *esh-put-char* *esh-pipe-char*
		    *esh-comment-char* *esh-include-char*

          read-next          	;read using the EpiShell command syntax
          set-symbol-syntax     ;set syntax of character to stand alone symbol
          set-default-syntax    ;set syntax of character to default
          special-char-p        ;true if character has special syntax
          reset-readtables      ;reset EpiShell readtables
	 ))
(unless (find-package 'esh-read) (make-package 'esh-read)))
(in-package esh-read)

(defvar *esh-take-char* #\<
        "Input redirection character for EpiShell" )

(defvar *esh-put-char* #\>
        "Output redirection character for EpiShell" )

(defvar *esh-pipe-char* #\|
        "Process pipe character for EpiShell" )

(defvar *esh-comment-char* #\;
        "EpiShell comment delimiter" )

(defvar *esh-include-char* #\@
        "File include character for EpiShell" )



;;; Reading for the EpiShell is done using a special readtable

(defun special-char-reader (stream char)
    (declare (ignore stream))
    char)

(defun special-sym-reader (stream char)
    (declare (ignore stream))
    (intern (string (char-upcase char))))
    

(defun comment-reader (stream char)
    (declare (ignore char))
    (do () ((char= (read-char stream nil #\newline t) #\newline)))
    #\newline)

(defun make-basic-table () 
    (let ((newtable (copy-readtable nil)))
        (if *esh-include-char*
            (set-macro-character *esh-include-char* #'special-char-reader 
							    nil newtable))
        (set-syntax-from-char #\; #\a newtable newtable)
        (if *esh-comment-char*
            (set-macro-character *esh-comment-char* #'comment-reader nil 
                                                                     newtable))
        (set-macro-character #\( #'special-char-reader nil newtable)
        (set-macro-character #\) #'special-char-reader nil newtable)
        (set-macro-character #\' #'special-char-reader nil newtable)
        (set-macro-character #\! #'special-sym-reader nil newtable)
        (set-macro-character #\newline #'special-char-reader nil newtable)
        newtable )
)

(defvar *esh-basic-readtable* (make-basic-table)
    "EpiShell basic readtable")


;;; The EpiShell command readtable includes a special reader for the
;;; put and take characters.
;;;
(defun make-cmd-table ()
    (let ((newtable (copy-readtable *esh-basic-readtable*)))
        (if *esh-take-char*
            (set-macro-character *esh-take-char* #'special-char-reader 
							    nil newtable))
        (if *esh-put-char*
            (set-macro-character *esh-put-char* #'special-char-reader 
							    nil newtable))
        (if *esh-pipe-char*
            (set-macro-character *esh-pipe-char* #'special-char-reader 
							    nil newtable))
        newtable)
)

(defvar *esh-cmd-readtable* (make-cmd-table) 
    "Readtable when reading EpiShell command files"
)

(defun set-symbol-syntax (char)
   (set-macro-character char #'special-sym-reader nil *esh-basic-readtable*)
   (set-macro-character char #'special-sym-reader nil *esh-cmd-readtable*)
)

(defun set-default-syntax (char)
   (set-syntax-from-char char char *esh-basic-readtable* nil)
   (set-syntax-from-char char char *esh-cmd-readtable* nil)
)
    
(defun special-char-p (char)
    (if (or (get-macro-character char *esh-basic-readtable*)
            (get-macro-character char *esh-cmd-readtable*)) t))
    
;;; Reset the readtables to their default, using the values of
;;; *esh-take-char*, *esh-put-char*, or *esh-comment-char*.
;;;
(defun reset-readtables ()
    (setq *esh-basic-readtable* (make-basic-table))
    (setq *esh-cmd-readtable* (make-cmd-table))
)


;;; read-next reads the next line in the input stream, using the command 
;;; readtable (but switching to the basic readtable whenever it reads a list).
;;; @filename constructs are expanded to include the contents of the
;;; named file.
;;;
;;;	input-stream is the stream to read from
;;;	open-files is a list of files that are currently open, and which
;;;		therefore cannot be included
;;;
;;;	Returns the contents of the line, as a list.
;;;
(defun read-next (&optional (input-stream *standard-input*) (open-files nil)
		 &aux (savetable *readtable*) )
    (unwind-protect
	(catch 'read-catcher 
	       (setq *readtable* *esh-cmd-readtable*)
	       (do ((objects 
    		       `(,(read-object input-stream *esh-basic-readtable*))
		       (cons (read-object input-stream *esh-basic-readtable*)
			     objects)))
		   ((or (eq (car objects) esh-eof)
			(not (listen input-stream))
			(eql (car objects) #\newline))
			 (if (and (not (eq (car objects) esh-eof))
				  (not (listen input-stream))
			          (not (eql (car objects) #\newline)))
			     (setq objects (cons #\newline objects)))
			 (expand-include (nreverse objects) 
                                                    *esh-basic-readtable* 
				 (cons (get-filename input-stream) 
				       open-files))
		    )))
	(setq *readtable* savetable))
)


;;; read-object reads the next object in the input stream.  If
;;; the object is a list and a list readtable is supplied, it will
;;; switch to that readtable while reading the list.
;;;
(defun read-object (input-stream list-readtable 
					&optional (in-list nil) &aux object)
    (setq object (read input-stream nil esh-eof))
    (do   nil 
          ((or in-list (not (eql object #\) ))))
       (context-msg input-stream "Ignoring an unmatched right parenthese.")
       (setq object (read input-stream nil esh-eof))) 
    (cond ((eql object #\') 
              (setq object (list 'user::quote 
                               (read-object input-stream list-readtable))))
          ((eql object #\( ) 
	      (if list-readtable
		(prog ((savetable *readtable*))
		    (setq *readtable* list-readtable)
		    (setq object (read-list input-stream ))
		    (setq *readtable* savetable))
		(setq object (read-list input-stream)))
	       ))
    object
)

;;; read-list reads objects from the input stream until it finds
;;; the end-of-list character.  End-of-file errors arre thrown to a
;;; read-catcher.
;;;
(defun read-list (input-stream &aux (list-object nil))
    (cond ((eq (do ((elt (read-object input-stream nil t) 
			 (read-object input-stream nil t)))
		   ((or (eql elt #\)) (eq elt esh-eof) ) 
                     elt)
                      ;; Ignore newlines in lists.  Unfortunately, we
                      ;; can't do this in the readtable, because then
                      ;; we lose newlines that appear immediately after 
                      ;; the end of a list
                      (unless (eql elt #\newline)
	                  (setq list-object (cons elt list-object)))) 
                esh-eof)
	       (context-msg input-stream 
					"End of file encountered in a list.")
               (throw 'read-catcher esh-read-error))
          (t (nreverse list-object)) ) ) 
	

;;; expand-include parses a list of objects and expands @ filename
;;; constructs into the contents of the named file.  Multiple occurences
;;; of the include character (e.g. @@filename) are handled properly. 
;;; Errors are thrown to a read-catcher.
;;;
;;;	object is the list of objects to expand
;;;	list-readtable is the readtable to switch to if the contents of
;;;		the included file(s) contain lists
;;;	file-stack is a list of files that are already open
;;;
;;;	Returns the expanded object.
;;;
(defun expand-include (object list-readtable file-stack &aux filename)
    (cond ((atom object)
	        object)
	  ((eql (car object) *esh-include-char*)
	 	(unless (cdr object) 
		    (context-msg (car file-stack) 
				"Include must be followed by a filename")
		    (throw 'read-catcher esh-read-error))
		(setq object (expand-include (cdr object) list-readtable 
					     file-stack))
		(unless (setq filename (check-filename (car object)
					       file-stack (car file-stack)))
		    (throw 'read-catcher esh-read-error))
		(nconc (expand-include 
			     (expand-file filename list-readtable
							 (car file-stack))
			     list-readtable (cons filename file-stack))
		       (cdr object)))

          ((eql (car object) #\newline)  ; ignore newlines in included files
             (expand-include (cdr object) list-readtable file-stack))

	  (t (cons (if list-readtable
		       (prog ((savetable *readtable*))
			     (setq *readtable* list-readtable)
			     (return 
				(prog1 (expand-include (car object) nil
							file-stack)
				       (setq *readtable* savetable))))
			(expand-include (car object) nil file-stack))
		    (expand-include (cdr object) 
						list-readtable file-stack))))
)


;;; expand-file opens and reads the contents of the named file, 
;;; returning them as a list.
;;;
(defun expand-file (filename list-readtable &optional (file-context nil))
    (with-open-file (instream filename :direction :input 
				       :if-does-not-exist nil)
	(unless instream 
	    (file-open-msg file-context filename)
	    (throw 'read-catcher esh-read-error))
	(do ((objects `(,(read-object instream list-readtable))
		      (cons (read-object instream list-readtable)
			    objects)))
	     ((eq (car objects) esh-eof) (nreverse (cdr objects)))))
)







