;       Last change    August 1993
(in-package number)


; the following routines are used to determine whether two literals are 
; incompatible or compatible

(defun quoted-number (arg)
  (if (and (listp arg)
	   (equal (cdr arg) 'quote)
	   (numberp (cadr arg)))
      t
      (numberp arg))
  )

; usable returns t if the arg given is a quoted number or exists in the given 
; number net; otherwise nil

(defun usable (net arg)
  (if (numberp arg)
      t
      (findnumber net arg)
      )
  )

; unifiable is used by the unify routine to determine if two constant arguments
; are "unifiable". They do not have to unify in the typical sense - it is
; enough if there is some relation between them. Since it is very time consuming
; to calculate this relation, a short cut is used whioch doesn't necessarily
; guarantee a relation - however, it will be tested during the entry/test
; phase anyway. Why do it twice.

(defun unifiable (arg1 arg2 &aux
			(net (determine-net *current-netname*)))
  (if (equal arg1 arg2)
      t
      (if (null net)
	  nil
	  (and (usable net arg1) (usable net arg2))
	  )
      )
  )

; determine enter decides which of two literals should be entered. If one
; is negated, the other is entered, if both are negated, neither can be 
; entered.
; If there is a choice, and one is a between predicate, enter that one, as
; it is harder to test with

(defun determine-enter (neg-p1 neg-p2)
  (if neg-p2
      (if neg-p1
	  nil
	  '1)
      '2)
  )


; classify-lits classifies the given literals, and returns a list
; of substitution orders and which literal to enter.
; Each item in the list is of the form (orderlist . number of lit to enter)
; Each item in orderlist is of the form 
;     (element number from 1st lit . element number from second)

(defun classify-lits (neg-p1 neg-p2 &aux
			     (which (determine-enter neg-p1 neg-p2))
			     ordlist)
  (unless (null which)
	 ; each literal has only 2 args which can be unified
	 ; only two possible unifications - same direction and reversed

	 (setq ordlist (list (cons '((1 . 1)(2 . 2)) which)
			     (cons '((1 . 2)(2 . 1)) which)))
	  )
  ordlist)

; create-residue creates a literal of the given neg pred ard args

(defun create-residue (neg pred &rest args &aux
			    (reln (find-symbol (symbol-name pred) 'epilog)))
  (if (or (null pred)
	  (null args)
	  (eq (car args) (cadr args)))
      t
      (make-prop neg :pred reln :args args)
      )
  )

(defun more-strict (reln)
  (case reln
	(lt= 'less-than)
	(gt= 'greater-than)
	(otherwise nil)
	)
  )

; diff-reln returns the difference between two relations , where reln1 subsumes
; reln2. 

(defun diff-reln (reln1 reln2)
  (if (equal reln2 'equal)
      (more-strict reln1)
      (if (equal reln1 'equal)
	  (more-strict reln2)
          'equal))
  )

; number-compare-preds returns the relation between the two predicates reln1 and reln2
; (equivalent, subsumes, etc)
(defun number-compare-preds (reln1 reln2)
  (case reln1
	(equal 
	 (case reln2
	       (equal 'equivalent)
	       ((lt= gt=) 'subsumed)
	       ((lt gt less-than greater-than) 'disjoint)
	       (otherwise 'unknown)))
	((lt less-than)
	 (case reln2
	       ((lt less-than) 'equivalent)
	       (lt= 'subsumed)
	       ((equal gt greater-than gt=) 'disjoint)
	       (otherwise 'unknown)))
	(lt=
	 (case reln2
	       ((lt less-than equal) 'subsumes)
	       (lt= 'equivalent)
	       (gt= 'intersects)
	       ((gt greater-than) 'disjoint)
	       (otherwise 'unknown)))
	((gt greater-than)
	 (case reln2
	       ((gt greater-than) 'equivalent)
	       (gt= 'subsumed)
	       ((equal lt less-than lt=) 'disjoint)
	       (otherwise 'unknown)))
	(gt=
	 (case reln2
	       ((gt greater-than equal) 'subsumes)
	       (gt= 'equivalent)
	       (lt= 'intersects)
	       ((lt less-than) 'disjoint)
	       (otherwise 'unknown)))
	(otherwise 'unknown)
	)
  )




; incompatible-test-pred tests the other literal, now that one has been entered.

(defun incompatible-test-pred (net negp reln args effort 
				   &optional (incompatible t) &aux
			      (a1 (car args))
			      (a2 (cadr args))
	                      (testreln (find-relation net a1 a2
					   *specialist-eval-effort*))
		              (creln (number-compare-preds reln testreln)))

	; compare this relation with the one between those
	; item sin the number graph. Depending on whether this literal
	; is negated or not, and this relationship, we can determine
	; a residue.

		 
        (declare (ignore effort))
	(if negp
	    (case creln
	          (equivalent (not incompatible))
	          (subsumes (not incompatible))
	          (subsumed  ;should be difference <= & = -> <
	           (create-residue nil
		             (diff-reln reln testreln) a1 a2))
		  (intersects (if (equal testreln 'equal)
				   nil
				   (create-residue nil
					      (more-strict testreln) a1 a2)))
		 (disjoint incompatible)
		 (otherwise t)
		 )
	    (case creln
	          (equivalent incompatible)
	          (subsumes incompatible) 
	          (subsumed (create-residue nil reln a1 a2))
	          (intersects (create-residue nil 'equal a1 a2))
	          (disjoint (not incompatible))
	          (otherwise t)
	          )
	     )
  )

; enter-to-test just ensuere that no negations get passed to enter

(defun enter-to-test (net negp pred arglist)
  (if negp
      nil
      (enter-number net pred arglist))
  )

; incompatible-add checks to make sure that the residues are not the same as the
; original literals (without substitutions). If so, a subgoal item is
; added, otherwise, nothing is done.

(defun incompatible-add (lit-1 lit-2 res1 res2 eval1 eval2 subs1 subs2
			       &optional (incompatible t) &aux
			       rr)
  (declare (ignore eval1 eval2))
  (unless (or (member lit-1 res1)
	      (member lit-1 res2)
	      (member lit-2 res1)
	      (member lit-2 res2))
          (when (traced-p 'epilog::number-test)
		    (print-line "    Number Specialist: Found " 
				(print-info lit-1 )
		     (if incompatible " incompatible" " compatible") "  with "
		     (print-info lit-2))
		(print-line "        with residues: "
			    (if (setq rr (append res1 res2)) rr "null"))
		(when (or subs1 subs2)
		    (print-line "        with substitutions: " 
			       (print-subs (append subs1 subs2)))))
	  (list 
	   (make-comparison-info :lit1 lit-1 :lit2 lit-2
			     :subs1 subs1 :subs2 subs2
			     :residue1 res1
			     :residue2 res2
			     ; evaluations
			     ) )
	  )
  )


; incompatible-enter-and-compare checkpoints the net, enters one literal, and
; tests the other
; The net is then retracted to its previous state

(defun incompatible-enter-and-compare (net lit-1 lit-2 entry eval1 eval2
				      neg-p1 reln1 newargs1
			              neg-p2 reln2 newargs2 subs1 subs2 effort 
				      &optional (incompatible t)
				      &aux
			              cp results resans)
  (unless (null entry)
   
  (setq cp (checkpoint "Number-CP"))
  (when (and (equal entry '2) (not (equal eval2 'no)))
 	(enter-to-test net neg-p2 reln2 newargs2)
	(setq resans (incompatible-test-pred net neg-p1 reln1 newargs1 effort
					     incompatible))
	(cond ((equal resans t)
	       nil)
	      (t (setq results (append results
		       (incompatible-add lit-1 lit-2 resans nil nil nil
				    subs1 subs2 incompatible))))
	      )
	)
  (when (and (equal entry '1) (not (equal eval1 'no)))
  	(enter-to-test net neg-p1 reln1 newargs1)
	(setq resans (incompatible-test-pred net neg-p2 reln2 newargs2 effort
					     incompatible))
	(cond ((equal resans t)
	       nil)
	      (t (setq results (append results
		       (incompatible-add lit-1 lit-2 nil resans nil nil
				    subs1 subs2 incompatible))))
	      )
	)
  (retract cp :remove t)
  )
  results
  )

; test-residue does the necessary substitutions, and then determines
; compatibility or incompatibility of the resulting literals

(defun test-residue (incompatible net lit-1 lit-2 ord-entry
			     neg-p1 reln1 arglist1
			     neg-p2 reln2 arglist2 effort &aux
			     a1 res
 	                     (ordlist (car ord-entry))
 	                     (entry (cdr ord-entry))
			     (newargs1 arglist1)
			     (newargs2 arglist2)
			     (eval1 'unknown)
			     (eval2 'unknown)
			     newlit1 newlit2
	                     ulist subs1 subs2)
  (unless (null ordlist)
	  ; do unification using specification given
          (setq ulist (unify-lists arglist1 arglist2 ordlist 'unifiable))

	  ; do substitutions in literals
          (unless (null ulist)
                  (setq subs1 (car ulist))
	          (setq subs2 (cdr ulist))
	          (setq newlit1 (normalize (substitute-prop lit-1 subs1) '+ nil nil))
	          (setq newlit2 (normalize (substitute-prop lit-2 subs2) '+ nil nil))
		  (when (or (equal (car (setq a1 (getp newlit1 'arguments)))
				   (cadr a1))
			    (equal (car (setq a1 (getp newlit2 'arguments)))
				   (cadr a1)))
			(setq newlit1 nil)
			(setq newlit2 nil))
		  ))

  (unless (or (null newlit1)
	      (null newlit2))
          
	  ; evaluate the arguments in case they've changed
	  (setq newargs1 (spec-eval-args newlit1))
	  (setq newargs2 (spec-eval-args newlit2))
	  (when (or (equal (car newargs1) (cadr newargs1))
		    (equal (car newargs2) (cadr newargs2)))
		(return-from test-residue nil))

	  ; evaluate the literals
	  (setq eval1 (eval-with-neg net neg-p1 reln1 newargs1 effort))
	  (setq eval2 (eval-with-neg net neg-p2 reln2 newargs2 effort))

	  ; make sure all substitutions known
          (when (equal eval1 'unknown)
                (setq subs1 (union subs1 
		                (arg-sub-list newargs1 
						   (getp newlit1 'arguments) 1)))
	        (setq newlit1 (normalize (substitute-prop lit-1 subs1) '+ nil nil))
	        )
          (when (equal eval2 'unknown)
                (setq subs2 (union subs2 
		              (arg-sub-list newargs2 
						 (getp newlit2 'arguments) 2)))
	        (setq newlit2 (normalize (substitute-prop lit-2 subs2) '+ nil nil))
	        )

	  ; first use evaluations if possible

	  (cond ((setq res ;if incompatible
                     (incompatible-with-eval lit-1 lit-2 eval1 eval2 subs1 subs2 
		                        (traced-p 'epilog::number-test)
		                        "    Number Specialist:"
					incompatible))
		 res)

		; if this gives no answer, enter one and test the other
		; ensure that we don't enter contradictory info into number graph

		((and (member eval1 '(YES NO)) (member eval2 '(YES NO)))
		 nil)

		((equal (if (equal entry 1) eval1 eval2) 'no) 
		 nil)

		(t (incompatible-enter-and-compare net 
				     lit-1 lit-2 entry eval1 eval2
			              neg-p1 reln1 newargs1
			              neg-p2 reln2 newargs2
			              subs1 subs2 effort incompatible)
		       )
		)
	  )
  )

; if arg is a constant, it must be of correct sort - if variable, must be
; of correct sort, or have no sort

(defun okarg (arg &aux
		  (s (find-sort arg)))
  (if (check-constant arg)
      (if (member s '(number integer real))
	  t
	  nil)
      (or (null s)
	  (if (member s '(number integer real))
	      t
	      nil))
      )
  )

(defun check-arg-sorts (arglist)
  (and (okarg (car arglist))
       (okarg (cadr arglist)))
  )

; incompatible-lits tries to determine whether lit1 and lit2 are incompatible
; it tries all possible substitutions to find out

(defun incompatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort 
			   &optional (incompatible t) &aux
		           (net (determine-net netname))
		           (reln1 (find-symbol (symbol-name pred1) 'number))
		           (reln2 (find-symbol (symbol-name pred2) 'number))
	                   (all-constants (not (or 
						(non-constant-arg arglist1)
				                (non-constant-arg arglist2)
						)))
			   )

  (when (and (check-arg-sorts arglist1)
	     (check-arg-sorts arglist2))

  ; see if arguments unifiable, and in what order

  ; enter one literal, then test the relationship for the
  ; arguments in the other one against the net relationship

  (when (traced-p 'epilog::number-test)
        (print-line "    Number Specialist: Trying to determine "
                  (if incompatible "incompatibility" "compatibility") 
		  " of " (print-info lit1) " and " (print-info lit2))
	(if (not (equal (number-subnet-identifier net) '/))   
	    (print-it "    in subnet " (number-subnet-identifier net)))
      )

  (if all-constants
      (let ((eval1 (eval-with-neg net neg-p1 reln1 arglist1 effort))
	    (eval2 (eval-with-neg net neg-p2 reln2 arglist2 effort)))
	   (cond ((incompatible-with-eval lit1 lit2 eval1 eval2 nil nil 
		                        (traced-p 'epilog::number-test)
		                        "    Number Specialist:" incompatible))
		 (t
                   (incompatible-enter-and-compare net 
		                              lit1 lit2 
		                              (determine-enter neg-p1 neg-p2)
					      eval1 eval2
			                      neg-p1 reln1 arglist1
			                      neg-p2 reln2 arglist2
			                      nil nil effort incompatible)
		   )))

      (let* ((ord-entry (classify-lits neg-p1 neg-p2)))
	    (if (null ord-entry)
	        nil
		(let (r)
		     (mapcar #'(lambda (o)
				       (setq r (append r
						 (test-residue 
						  incompatible
					 	  net lit1 lit2 o 
						  neg-p1 reln1 arglist1
					          neg-p2 reln2 arglist2 
						  effort))) )
				      ord-entry)
		     r)
		)
	    )
      ))
  )

; compatible-lits tries to determine whether lit1 and lit2 are compatible
; it tries all possible substitutions to find out

(defun compatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort )
  (incompatible-lits netname lit1 lit2 neg-p1 pred1 arglist1 neg-p2 pred2
		     arglist2 effort nil)
  )
