;       Last change    August 1993
(in-package number)

;; Entry routines

; set-min determines the actual minimum to use for a number and calls
; set-exact-min to set it

(defun set-min (nbr mingiven exact &aux
		    (nbrmin (epinumber-min nbr))
		    (nbrmax (epinumber-max nbr))
		    (min (calc-min nbr mingiven)))
  (unless (null nbr)
	  (if (and (not exact)
		   (equal (epinumber-ntype nbr) 'integer)
	           (equal mingiven min))
	      (set-exact-min nbr nbrmin nbrmax (+ min 1))
	      (set-exact-min nbr nbrmin nbrmax min)
	      ))
  )

; set-exact-min checks to make sure the given minimum better than the
; previous minimum before changing t. If the minimum is changed, the new
; minimum is propagated forward.
;
(defun set-exact-min (nbr nbrmin nbrmax min)
  (when (and (or (null nbrmin)
	         (> min nbrmin))
	     (or (null nbrmax)
	         (<= min nbrmax)))
	(if (traced-p 'epilog::number-entry)
	    (print-line "    Number specialist: Setting minimum of " 
			     (epinumber-name nbr)  " to " min))
        (change-field 'epinumber-min nbr min)
        (if (equal (epinumber-min nbr) (epinumber-max nbr))
	    (change-field 'epinumber-value nbr min))
	(propagate-min nbr)
	t
	)
  )

; set-max determines the actual maximum to use for a number and calls
; set-exact-max to set it

(defun set-max (nbr maxgiven exact &aux
		    (nbrmin (epinumber-min nbr))
		    (nbrmax (epinumber-max nbr))
		    (max (calc-max nbr maxgiven)))
  (unless (null nbr)
	  (if (and (not exact)
		   (equal (epinumber-ntype nbr) 'integer)
		   (equal maxgiven max))
	      (set-exact-max nbr nbrmin nbrmax (- max 1))
	      (set-exact-max nbr nbrmin nbrmax max)
	      ))
  )

; set-exact-max checks to make sure the given maximum better than the
; previous maximum before changing t. If the maximum is changed, the new
; maximum is propagated backward.
;
(defun set-exact-max (nbr nbrmin nbrmax max)
  (when (and (or (null nbrmax)
	         (< max nbrmax))
	     (or (null nbrmin)
  	         (>= max nbrmin)))
	(if (traced-p 'epilog::number-entry)
	    (print-line "    Number specialist: Setting maximum of " 
			     (epinumber-name nbr)  " to " max))
        (change-field 'epinumber-max nbr max)
        (if (equal (epinumber-min nbr) (epinumber-max nbr))
	    (change-field 'epinumber-value nbr max))
	(propagate-max nbr)
	t
	)
  )

; set-value sets both the minimum and maximum of a number and propagates
; it forward and backward 

(defun set-value (nbr valgiven &aux
		      (nbrmax (epinumber-max nbr))
		      (nbrmin (epinumber-min nbr))
		      (val (if (equal (epinumber-ntype nbr) 'integer)
			       (round valgiven)
			       (float valgiven))))
  (when (and (null (epinumber-value nbr))
	     (or (null nbrmax)
	         (<= val nbrmax))
	     (or (null nbrmin)
	         (>= val nbrmin)))
	(if (traced-p 'epilog::number-entry)
	    (print-line "    Number specialist: Setting value of " 
			     (epinumber-name nbr) " to " val))
        (change-field 'epinumber-value nbr val)
        (change-field 'epinumber-min nbr val)
        (change-field 'epinumber-max nbr val)
	(propagate-min nbr)
	(propagate-max nbr)
	t
	)
  )

; reverse-reln determines the inverse of the given relation

(defun reverse-reln (reln)
  (case reln 
	(equal 'equal)
	((less-than lt) 'greater-than)
	((greater-than gt) 'less-than)
	(lt= 'gt=)
	(gt= 'lt=)
	(t 'unknown)
	)
  )

; enter-reln enters a numeric relation between a named number and a number

(defun enter-reln (net a1 reln a2 &aux
		       (cname a1)
		       (nbr (getnumber net cname (find-sort a1)))
		       (value a2))
  (unless (null reln)
	  (case reln
		(equal (set-value nbr value))
		((less-than lt) (set-max nbr value nil))
		((greater-than gt) (set-min nbr value nil))
		(lt= (set-max nbr value t))
		(gt= (set-min nbr value t))
		)
	  )
  )

; propagate-min propagates a minimum forward. At each stage, calc-min is 
; used to determine what the minimum proagated will be (in case of integers
; and reals being propagated to).

(defun propagate-min (num &aux
			  (min (epinumber-min num))
			  (links (append (epinumber-next num) 
					 (epinumber-strict-next num))))
  (unless (null min)
  (dolist (lnum links)
	  (let* ((l (findnumber current-net lnum))
		 (newmin (calc-min l min)))
	  (when (and (or (null (epinumber-min l))
		         (> newmin (epinumber-min l)))
	             (or (null (epinumber-max l))
	                 (<= newmin (epinumber-max l))))
	        (if (traced-p 'epilog::number-prop)
	            (print-line
		     "    Number specialist: Propagating minimum " newmin  
			     " to " (epinumber-name l)))
		(spec-changed-concept (epinumber-name l))
	        (change-field 'epinumber-min l newmin)

		; if min and max are equal, set value
                (if (equal (epinumber-min l) (epinumber-max l))
	            (change-field 'epinumber-value l min))

		; propagate new value
		(propagate-min l)
		))
	  ))
  )

; calc-min determines what the minimum of a number should be - if the
; given minimum is an integer, and the number it is being sent to is alos
; an integer, no changes are necessary. If the propagated number is not an
; integer, its ceiling is used. For reals, the minimum is floated

(defun calc-min (num min)
  (if (equal (epinumber-ntype num) 'integer)
      (if (integerp min)
	  min
	  (ceiling min))
      ; else real
      (float min))
  )

; calc-max determines what the maximum of a number should be - if the
; given maximum is an integer, and the number it is being sent to is alos
; an integer, no changes are necessary. If the propagated number is not an
; integer, its floor is used. For reals, the maximum is floated

(defun calc-max (num max)
  (if (equal (epinumber-ntype num) 'integer)
      (if (integerp max)
	  max
	  (floor max))
      ; else real
      (float max))
  )

; propagate-max propagates a maximum backward. At each stage, calc-max is 
; used to determine what the maximum propagated will be (in case of integers
; and reals being propagated to).

(defun propagate-max (num &aux
			  (max (epinumber-max num))
			  (links (append (epinumber-prev num) 
					 (epinumber-strict-prev num))))
  (unless (null max)
  (dolist (lnum links)
	  (let* ((l (findnumber current-net lnum))
		 (newmax (calc-max l max)))
	  (when (and (or (null (epinumber-max l))
		         (< newmax (epinumber-max l)))
	             (or (null (epinumber-min l))
	                 (>= newmax (epinumber-min l))))
	        (if (traced-p 'epilog::number-prop)
	            (print-line
		     "    Number specialist: Propagating maximum " newmax  
			     " to " (epinumber-name l)))

		; if min and max are equal, set value 

		(spec-changed-concept (epinumber-name l))
	        (change-field 'epinumber-max l newmax)
                (if (equal (epinumber-min l) (epinumber-max l))
	            (change-field 'epinumber-value l max))

		; propagate new value

		(propagate-max l)
		))
	  ))
  )

  ; copy links from nm2 to nm1
(defun copy (net nm1 nm2 &aux
		 (fs (epinumber-strict-next nm2))
		 (fns (epinumber-next nm2))
		 (bs (epinumber-strict-prev nm2))
		 (bns (epinumber-prev nm2))
		 (fs1 (epinumber-strict-next nm1))
		 (fns1 (epinumber-next nm1))
		 (bs1 (epinumber-strict-prev nm1))
		 (bns1 (epinumber-prev nm1))
		 (name1 (epinumber-name nm1))
		 lname)
  (dolist (l fs)
	  (unless (or (equal name1 l)
		      (equal name1 (epinumber-name (findnumber net l))))
		  (pushnew l fs1)))
  (dolist (l fns)
	  (setq lname (epinumber-name (findnumber net l)))
	  (unless (or (equal name1 l)
		      (equal name1 lname)
		      (member l fs1)
		      (member lname fs1))
	          (pushnew l fns1)))
  (dolist (l bs)
	  (unless (or (equal name1 l)
		      (equal name1 (epinumber-name (findnumber net l))))
	          (pushnew l bs1)))
  (dolist (l bns)
	  (setq lname (epinumber-name (findnumber net l)))
	  (unless (or (equal name1 l)
		      (equal name1 lname)
		      (member l bs1)
		      (member lname bs1))
	          (pushnew l bns1)))

  ; remove any ptrs from 1 to 2
  (setq fns (remove (epinumber-name nm2) fns1 :test #'equal))
  (setq fs (remove (epinumber-name nm2) fs1 :test #'equal))
  (setq bns (remove (epinumber-name nm2) bns1 :test #'equal))
  (setq bs (remove (epinumber-name nm2) bs1 :test #'equal))

  (change-field 'epinumber-next nm1 fns)
  (change-field 'epinumber-strict-next nm1 fs)
  (change-field 'epinumber-prev nm1 bns)
  (change-field 'epinumber-strict-prev nm1 bs)
  )

(defun collapse (net num1 num2)
  (copy net num1 num2)
  (if (traced-p 'epilog::number-entry)
      (print-line "    Number specialist: Collapsing points " 
		       (epinumber-name num1) " and " (epinumber-name num2)))
  (change-hash (epinumber-name num2) 
	       (number-subnet-numbers net) num1) ; set nm2 to nm1
  (propagate-min num1)
  (propagate-max num1)
  )

(defun enter-equal (net arg1 arg2 nm1 nm2)
  (if (equal arg2 (epinumber-name nm2))
      (collapse net nm1 nm2)
      (let ()
	   (enter-equal net arg1 (epinumber-name nm2) nm1 
		(getnumber net (epinumber-name nm2) (epinumber-ntype nm2)))
	   (if (traced-p 'epilog::number-entry)
	       (print-line "    Number specialist: Collapsing points " 
				arg1 " and " arg2))
	   (change-hash (epinumber-name nm2) 
			(number-subnet-numbers net) nm1) ; set nm2 to nm1
	   )
      )
  )

; enter-link enters a relation between two named numbers 
; minima and maxima are propagated after to ensure that each number has the 
; best bounds possible

(defun enter-link (net a1 reln a2 &aux
		       (arg1 a1)
		       (arg2 a2)
		       (s1 (find-sort a1))
		       (s2 (find-sort a2))
		       (nm1 (getnumber net arg1 s1))
		       (nm2 (getnumber net arg2 s2)))
  (case reln 
	(equal
         ; If two numbers are set equal, all links from one are copied to
         ; the ohter, and then the two points set to the same number.
	 (enter-equal net arg1 arg2 nm1 nm2)
	 )

	((less-than lt)
	 ; a strict link added
	 (change-field 'epinumber-strict-next nm1 (append (epinumber-strict-next nm1) 
						       (list (epinumber-name nm2))))
	 (change-field 'epinumber-strict-prev nm2 (append (epinumber-strict-prev nm2) 
						       (list (epinumber-name nm1))))
	 (propagate-min nm1)
	 (propagate-max nm2))

	(lt=
	 ; a non-strict link added
	 (change-field 'epinumber-next nm1 (append (epinumber-next nm1) (list (epinumber-name nm2))))
	 (change-field 'epinumber-prev nm2 (append (epinumber-prev nm2) (list (epinumber-name nm1))))
	 (propagate-min nm1)
	 (propagate-max nm2))

	((greater-than gt)
	 ; a strict link added
	 (change-field 'epinumber-strict-next nm2 (append (epinumber-strict-next nm2) 
						       (list (epinumber-name nm1))))
	 (change-field 'epinumber-strict-prev nm1 (append (epinumber-strict-prev nm1) 
						       (list (epinumber-name nm2))))
	 (propagate-min nm2)
	 (propagate-max nm1))

	(gt=
	 ; a non-strict link added
	 (change-field 'epinumber-next nm2 (append (epinumber-next nm2) (list (epinumber-name nm1))))
	 (change-field 'epinumber-prev nm1 (append (epinumber-prev nm1) (list (epinumber-name nm2))))
	 (propagate-min nm2)
	 (propagate-max nm1))
	)
  t
  )

; enter takes a given literal and enters it into the number graph representation
; only positive (non-negated) literals are entered, and each must have exactly 
; two number arguments.

(defun enter (netname lit negp pred arglist &aux
		      (reln (find-symbol (symbol-name pred) 'number))
		      (net (determine-net netname))
		      (a1 (car arglist))
		      (a2 (cadr arglist))
		      (a3 (caddr arglist)))
  (declare (ignore lit))
  (when (and (not negp)
	     (null a3)
	     (number-arg a1)
	     (number-arg a2))
	(when (traced-p 'epilog::number-entry)
	      (print-line "    Number specialist: Entering " 
			     (print-info a1) " " reln " " (print-info a2))
	      (if (not (equal (number-subnet-identifier net) '/)) 
			      (print-it "    into subnet " 
					(number-subnet-identifier net))) 
	      )
	(enter-number net reln arglist)
	)
  )

; enter-number checks to see what arguments have been given and calls the
; appropriate routine to add a link or change a bound accordingly.

(defun enter-number (net reln arglist &aux
			 (a1 (car arglist))
			 (a2 (cadr arglist)))
  ; can only enter when 1 is named constant, other number
 (if (symbolp a1)
     (if (symbolp a2)
	 (enter-link net a1 reln a2)
	 (enter-reln net a1 reln a2))
     (if (symbolp a2)
	 (enter-reln net a2 (reverse-reln reln) a1)
	 nil)   ; both numbers
	 )
  )
