;       Last change    August 1993
(in-package epilog)

;; Routines to do input-driven inferencing - including inference with mp's
;; Some extra restrictions hold for conclusions drawn from mp's (no "meta"
;; variables), and they may only be applied in the forward direction, where
;; regular knowledge rules can be applied either forward or backward
;; (contrapositive).

(tweakable '*forward-effort* 0
  "Effort level to verify antecedents with" '(float * *))
(tweakable '*forward-full* nil
  "Indicates whether only full instantiations allowed" 'boolean)
(tweakable '*look-up-members* t
  "Indicates whether or not to look up set members during inference" 'boolean)
(tweakable '*minimum-trace-items* nil
  "List of specific formulas to trace during forward-inference process")

; Forward trace stuff - using knowledge rules
(traceable 'forward "Shows forward inferences made")
(traceable 'forward-qa "Shows mini-question attempts during forward inference")
(traceable 'forward-test "Shows comparisons being made in forward inference")
(traceable 'forward-class 
	    "Shows classifications being used in forward inferencing")
(traceable 'forward-unify 
	    "Shows unifications being used in forward inferencing")
(traceable 'forward-access 
  "Shows accesses of formulas during forward inferencing")
(traceable 'infer-details 
  "Shows detail of forward inference decision")
(traceable 'forward-eval 
	    "Shows evaluations of literals during forward inference")
(traceable 'forward-all "Shows all forward inferencing attempts"
  '(forward forward-access forward-test forward-qa forward-eval forward-class))
(traceable 'forward-int "Shows interesting parts of forward inferencing"
  '(forward forward-eval forward-access forward-test forward-qa))
(traceable 'forward-min "Shows minimum parts of forward inferencing"
  '(forward forward-eval))

; MP trace stuff - inference using mps
(traceable 'mp-test "Shows comparisons being made in mp inference")
(traceable 'mp-class "Shows classifications being used in mp inferencing")
(traceable 'mp-access "Shows accesses of formulas during mp inferencing")
(traceable 'simplification-schema "Shows replacement of wff by simplification schemas")
(traceable 'mp-eval "Shows evaluations of literals during mp inference")
(traceable 'mp-all "Shows all mp inferencing attempts"
  '(forward forward-unify mp-access mp-test mp-eval mp-class 
	    simplification-schema))
(traceable 'mp-int "Shows interesting parts of mp inferencing"
  '(forward mp-eval mp-test mp-access simplification-schema))
(traceable 'mp-min "Shows minimum parts of mp inferencing"
  '(forward mp-eval simplification-schema))

; Entry traces - time, evaluations, also forward and mp stuff
(traceable 'entry-time "Keeps track of how much time each entry takes")
(traceable 'entry "Shows results of formula entry" '(entry-time forward))
(traceable 'entry-all "Everything about forward inferencing - mp too"
  '(entry forward-all mp-all entry-eval))
(traceable 'entry-int "Interesting stuff about forward inferencing - mp too"
  '(entry forward-int mp-int entry-eval))
(traceable 'entry-min "Minimum stuff about forward inferencing - mp too"
  '(entry forward-min mp-min entry-eval))

(defvar *key-verify* nil)
(defvar *ok-to-verify* nil)

(defun input-driven-inferences (item rules-so-far assumptions &optional pr-prop &aux
				     (+keys (getp item '+keys))
				     (-keys (getp item '-keys))
				     (+other (getp item '+other))
				     (-other (getp item '-other))
				     results used-classes already-wff 
				     lit context classes
				     (proof-prop (cond (pr-prop)
						       (t (if (getp item 'PROOF) 'PROOF
						     (if (getp item 'DISPROOF)
							 'DISPROOF)))))
				    (lit-classes (getp item 'lit-classes)))
  (when (null lit-classes)
	(classify item)
	(setq lit-classes (getp item 'lit-classes))
	)
  (dolist (cl lit-classes)
    (setq lit (car cl))
    (setq context (cadr cl))
    ; episode description stuff better suited to mp's
    (setq classes (remove-if #'(lambda (a) (equal (cdr a) 'tp.description))
			     (caddr cl)))
    (setq used-classes nil)
    (setq already-wff nil)

    (do ((c (car classes) (car classes)))
	((null c))
	(setq classes (cdr classes))
	(unless (member c used-classes :test 'equal)
		(push c used-classes)
		(if (traced-p 'forward-class)
		    (print-line "Looking for wffs for " (getp item 'print) 
				" under class " c))
		; lookup wffs
		(dolist (w (get-class c))
			(when (and (not (member w already-wff :test 'equal))
				   (not (member w rules-so-far :test 'equal))
				   (if (getp w 'assumption)
				       (member w assumptions) t)
				   (ok-infer item w))
			  (push w already-wff)
			  (setq results (append results 
						(try-with-wff nil item lit context w +keys -keys
							      +other -other proof-prop rules-so-far
							      assumptions)))
			  )
		  )
	  
		(when (and *simplification-schemas* results) (return))

		;; add classifications for parents of these now
		(if (getp item 'conditional)
		    (setq classes (union classes (sub-class c ) :test 'equal))
		    (setq classes (union classes (super-class c ) :test 'equal))
		    )
		))
	  )
  (remprop item 'lit-classes)
  results)

(defun try-with-wff (mp item lit context w +keys -keys +other -other proof-prop
			  rules-so-far assumptions &aux
			  results
			  (w+keys (getp w '+keys))
			  (w-keys (getp w '-keys))
			  (w+trigger (getp w '+trigger))
			  (w-trigger (getp w '-trigger))
			  (w+other (getp w '+other))
			  (w-other (getp w '-other))
			  (save-trace *traced-items*))
  (when (and *minimum-trace-items* (member w *minimum-trace-items*))
	(trace-item 'mp-int 'forward-int 'meta-eval))
  (when (if mp
	    (or (traced-p 'mp-test) (traced-p 'mp-access))
	    (or (traced-p 'forward-test) (traced-p 'forward-access)))
	(print-line "Trying " (if mp "mp " "") "inference with " 
		    item ": " (getp item 'print) *nl* *spaces*
		    "       and " w ": " (getp w 'print)) )
	
  (setq results (try-wff-keys mp item lit context +keys -keys +other -other
			      proof-prop w w+trigger w-trigger w+keys w-keys
			      w+other w-other rules-so-far assumptions))
  (unless *simplification-schemas* ; was and *ss* results
	  (if (getp w 'equivalence)
	      (setq results (append results
		       (try-wff-keys mp item lit context +keys -keys +other
				     -other proof-prop w w-trigger w+trigger
				     w-keys w+keys w-other w+other 
				     rules-so-far assumptions)))
		 ) )
  (unless *simplification-schemas* ; was and *ss* results
	  (if (getp item 'equivalence)
	      (setq results (append results
		       (try-wff-keys mp item lit (if (equal context '+) '- '+)
				     -keys +keys -other +other proof-prop 
				     w w+trigger w-trigger w+keys w-keys
				     w+other w-other rules-so-far assumptions)))
		 ) )
  (setq *traced-items* save-trace)
  results)

(defun try-wff-keys (mp item lit context +keys -keys +other -other proof-prop
			w w+trigger w-trigger w+keys w-keys w+other w-other 
			rules-so-far assumptions &aux results)

  ;; currently mp's are applied in the forward direction only; regular
  ;; knowledge rules can be applied in either direction
	   
  ; prevent contrapositive if mp or we're dealing with a type literal
  (unless (or mp (type-lit-p item))
  (dolist (k w+trigger)
	  ; if +trigger is a type predication, don't try for contrapositive
	  (unless (and (type-lit-p k)
		       (if (getp item 'conditional) (equal context '+)))
	  (setq results (append results
		(if (getp item 'conditional)
		    (infer-with mp w k '+ item lit context 
				(remove lit +keys) (remove lit -keys) 
				+other -other  1 proof-prop nil rules-so-far
				assumptions)
		    (infer-with mp item lit context w k '+ 
				(remove k w+keys) w-keys w+other w-other  
				1 proof-prop nil rules-so-far assumptions)
		    ))))
	  (when (and *simplification-schemas* results) (return))
	  ) )
  (dolist (k w-trigger)
	  (setq results (append results
	        (if (and (not mp) (getp item 'conditional))
		    (infer-with mp w k '- item lit context 
				(remove lit +keys) (remove lit -keys) 
				+other -other  1 proof-prop nil rules-so-far
				assumptions)
		    (infer-with mp item lit context w k '- 
				w+keys (remove k w-keys) w+other w-other  
				1 proof-prop nil rules-so-far assumptions)
		    )))
	  (when (and *simplification-schemas* results) (return))
	  )
  results)

(defun key-list (-list +list)
  (append (mapcar #'(lambda (a) (list a '- a)) -list)
	  (mapcar #'(lambda (a) (list a '+ a)) +list)) )
		
(defun infer-with (mp wff lit lit-context rule key key-context
		       +keys -keys +other -other prob proof-prop parents 
		       rules-so-far assumptions
		       &aux results)
  (when (if mp (traced-p 'mp-test) (traced-p 'forward-test))
	(print-line "    Comparing " (getp key 'print) " " key-context " with " 
		    (getp lit 'print) " " lit-context))
  ; try to match lit with key; if successful, try for rest of keys too
  (dolist (res (compare-wffs key lit (equal key-context '-) 
			     :prefer 2 
			     :traced (if mp (traced-p 'mp-test)
				 (traced-p 'forward-test)) 
			     :using-mp mp
			     :header "      "
			     :entering t))
	  (unless (or (comparison-info-residue1 res)
		      (comparison-info-residue2 res))
	  ; got a match
	  ; call to get rest of keys matched
	  (setq results (append results
             (infer-rest mp (list wff lit) wff rule key key key-context
			; (equal lit-context key-context) 
			 res 
			 (key-list (remove key -keys) (remove key +keys))
			 (key-list -other +other) nil prob proof-prop
			 parents rules-so-far assumptions)
				)) )
	  )
  results)

(defun infer-rest (mp orig wff rule key okey ctx res keys other ukeys prob 
			proof-prop parents rules-so-far  assumptions
			&optional subs1 subs2 
		       &aux p lit1 residue1 residue2 results nsubs1 nsubs2
		       mem set key-info orig-key context wffs members set-sort)
  (when res
	(setq lit1 (comparison-info-lit1 res))
	(setq nsubs1 (comparison-info-subs1 res))
	(setq nsubs2 (comparison-info-subs2 res))
	(setq residue1 (comparison-info-residue1 res))
	(setq residue2 (comparison-info-residue2 res))
	(unless (or residue1 residue2)
          (setq subs1 (append subs1 nsubs1 
			      (list (list okey (if (equal ctx '+) 'NO 'YES)))))
          (setq subs2 (append subs2 nsubs2))
	  (setq keys (remove-if #'(lambda (a) (equal (car a) lit1)) keys))
	  (unless (or (member wff parents) (member wff (getp rule 'parents)))
	  (setq prob (combine-probs prob 
		 (cond ((equal (setq p (getp wff 'prob)) 1) (list wff))
		       ((and (numberp p) (or (getp wff 'support) (getp wff 'contradict))) 
			(list wff))
			  (t p))))  )
	  (setq parents (union parents (list wff))) 
	  ))
  (when nsubs1 (setq keys (do-subs keys nsubs1)))
  (setq keys (remove-if #'(lambda (a)
			  (when (not (var-present (car a))) 
			        (let () (push (append a (list t)) other) t)
			      ))
			keys))
  (when (null keys)
	(return-from infer-rest
		     (setq results
		     (finish-infer mp orig rule other ukeys subs1 subs2 
					   prob proof-prop
					   parents rules-so-far assumptions))))

  (setq keys (order-keys keys))

  (setq key-info (car keys))
  (setq key (car key-info))
  (setq context (cadr key-info))
  (setq orig-key (caddr key-info))

  (when (if mp (traced-p 'mp-access) (traced-p 'forward-access) )
  (print-line "    Looking up formulas to match " context " key: " (getp key 'print) 
	      *nl* *spaces* "    in " rule ": " (getp rule 'print) 
	      *nl* *spaces* "    for wff " wff ": " (getp wff 'print)))
  ; look up wffs to compare against key
  ; if no formulas found, go on to next key
 
  (setq wffs (remove-duplicates (apply #'append 
    (mapcar #'(lambda (cl) (get-class cl)) (classify key)))))
  ; if key is of form ... member-of .., and set specialist active, get
  ; set specialist to return some members - all these can do is add
  ; substitutions
  (when (and *look-up-members* (member 'set-specialist *active-specialists*)
	     (equal (getp key 'pred) 'member-of)
	     (epi-variable (setq mem (car (getp key 'arguments))))
	     (not (epi-variable (setq set (cadr (getp key 'arguments)))))
	     (or (null (setq set-sort (find-sort set)))
		 (equal set-sort 'set)))
	(setq members (cddr (getp (apply (find-symbol "SET-MEMBERS" 'set) set) 'print))))
  (remprop 'lit-classes key)
  (when (and (null members) (null wffs)) (return-from infer-rest
     (infer-rest mp orig nil rule nil nil nil nil 
		 (remove-if #'(lambda (a) (equal (car a) key)) keys)
		 other (append ukeys (list key-info)) prob proof-prop parents 
		 rules-so-far assumptions
		 subs1 subs2)) )


  (dolist (m members)
	  (when (traced-p 'mp-access)
		(print-line "    Trying with set member " (print-info m) 
			    " for " mem))
	  (setq res (make-comparison-info :subs1 (list (list mem m))))
	  (setq results (append results
			(infer-rest mp orig nil rule key orig-key context ;cx
				       res keys other ukeys
				       prob proof-prop parents rules-so-far
				       assumptions subs1 subs2)))
	  )


  ; only accept simple formulas
  (dolist (w wffs)
	  (when (if (getp w 'assumption) (member w assumptions) t)
	  (when (simple-wff w)
		(when (if mp (traced-p 'mp-test) (traced-p 'forward-test))
		      (print-line "    Comparing " (getp w 'print)  " with key "
				  (getp key 'print)))
		(dolist (nres (compare-wffs key (car (getp w '+keys)) 
					    (equal context '-)
					    :prefer 1
					    :traced (if mp (traced-p 'mp-test)
						(traced-p 'forward-test)) 
					   :using-mp mp
					   :header "      "
					   :entering t))
			(setq results (append results
                           (infer-rest mp orig w rule key orig-key context ;cx
				       nres keys other ukeys
				       prob proof-prop parents rules-so-far
				       assumptions subs1 subs2)))
			)) ) )
  ; flag flag changes
  (if (or (null results) (type-lit-p key)
	  (> (list-length (getp rule '-keys)) 
	     (list-length (getp rule '-trigger)))
	  (> (list-length (getp rule '+keys)) 
	     (list-length (getp rule '+trigger)))
	  )
      (append results
         (infer-rest mp orig nil rule nil nil nil nil 
		 (remove-if #'(lambda (a) (equal (car a) key)) keys)
		 other ;(if (null results) 
			    (append ukeys (list key-info)) ;ukeys) 
		 prob proof-prop 
		 parents rules-so-far assumptions subs1 subs2))
      results)
  )

			
(defun warn-prob (wff)
  (print-line " WARNING: Probability calculation involving " wff ": " 
	      (getp wff 'print) " may be inaccurate")
  )

(defun simple-wff (wff)
  (equal (car (getp wff '+keys)) wff) )

(defun finish-infer (mp orig rule other ukeys subs1 subs2 prob proof-prop
			  parents rules-so-far assumptions
			  &aux res wff lit p nres uk
			  (rank 0) (input-index nil) olit
			  (vars (mapcar 'car (getp rule 'epilib::types)))
			  ores remvars pr
			  (+miss 0) (-miss 0) us
			  (+omiss 0) (-omiss 0)
			  (op *ordered-pieces*)
			  (oldpieces *pieces*))
  (dolist (u ukeys)
	  (setq uk (car u))
	  (cond ((and (symbolp (setq pr (getp uk 'pred)))
		      (member 'meta-specialist (getp pr 'specialists) ))
		 (setq other (append other (list u)))
		 )
		(t (setq us (append us (list u))))))
  (setq ukeys us)
  ; look up others if possible
  (dolist (o other)
	  (setq *interim* t)
	  (setq olit (normalize (substitute-prop (car o) subs1) '+ nil nil))
	  (setq *interim* nil)
	  (setq ores nil)
	  (when (if (not (var-present olit)) t 
		    (if (symbolp (setq pr (getp olit 'pred))) 
			(member 'meta-specialist (getp pr 'specialists) )
			nil)
		    )
		(if (if mp (traced-p 'mp-eval) (traced-p 'forward-eval))
		      (print-line "    Input-driven inference: Verifying: " 
				  (getp olit 'print)))
		(setq *ok-to-verify* t)
		(if (and (> *forward-effort* 0) (equal (cadr o) '-) 
			 (not proof-prop)
			 (not (equal (cadddr o) t)))
		    ; way to ensure that result only with prob 1?
		    (setq ores (key-verify olit)) 
		    (setq ores (verify olit proof-prop (if mp 2 1) nil assumptions))) 
		(setq *ok-to-verify* nil)
		)
	  (unless ores (if (equal (cadr o) '+) (incf +omiss) (incf -omiss)))
	  (when ores
		(if (if mp (traced-p 'mp-eval) (traced-p 'forward-eval))
		      (print-it "      -> " (car ores)))
		(if (and (equal (car ores) 'NO) (equal (cadr o) '+) 
			 (type-lit-p olit))
		    (return-from finish-infer nil))
		(setq parents (union parents (cadr ores)))
		(setq prob (combine-probs prob (caddr ores)))
		(setq subs1 (append subs1 (list (list (caddr o) (car ores)))))
		)
	  )
		  
  (dolist (u ukeys) (if (equal (cadr u) '+) (incf +miss) (incf -miss)))

  (setq remvars (set-difference vars (mapcar 'car subs1) :test 'equal))
  (setq remvars (remove-if-not #'(lambda (a) (getp a 'matchable)) remvars))
  (when (traced-p 'infer-details)
	(print-line "    Input-driven inference: Deciding on " 
		    (if mp "meaning postulate " "") "inference with ")
	(print-line "     " (car orig) ": " (print-info (car orig))) 
	(print-line "       and ")
	(print-line "     " rule ": " (print-info rule)) 
	(print-line "      Remaining variables: " remvars) 
	(print-line "      Original variables: " vars) 
	(print-line "      Substitutions: ") 
	(print-trace-subs subs1) (print-it " in " (car orig) )
	(print-line "                and  ") 
	(print-trace-subs subs2) (print-it " in " rule) 
	(print-line "      Unsatisfied trigger keys - negative: " -miss) 
	(print-line "                               - positive: " +miss) 
	(print-line "      Unsatisfied other keys - negative: " -omiss) 
	(print-line "                             - positive: " +omiss )
	)

  (setq +miss (+ +miss +omiss))
  (setq -miss (+ -miss -omiss))
  (cond (mp (if (or (member-if #'(lambda (a) 
		  (member (find-sort a) '(meta-entity e-term pred wff operator term quantifier)))
			   remvars)
		    (> -miss 0))
		(return-from finish-infer nil))
	    (if (and (> +miss 0) (> -miss 0) (= (length remvars) 0))
		(return-from finish-infer nil))  )
	(t (if (and *forward-full* (> -miss 0) (> (length remvars) 0))
	       (return-from finish-infer nil))   ))

  (setq subs1 (mapcar #'(lambda (a) (if (and (symbolp (cadr a))
					     (equal (getp (cadr a) 'entity-type)
						    'wff))
					(list (car a)
					      (substitute-prop (cadr a) subs2)
					      (caddr a))
					a)) subs1))
  (setq subs2 (mapcar #'(lambda (a) (if (and (symbolp (cadr a))
					     (equal (getp (cadr a) 'entity-type)
						    'wff))
					(list (car a)
					      (substitute-prop (cadr a) subs1)
					      (caddr a))
					a)) subs2))

  ; do necessary subs2 on rule
  (setq res (substitute-prop rule subs1))
  (setq prob (combine-probs prob *subst-prob*))
  ; if orig, replace lit in orig with result of subs and do subs1 on it
  (when orig
	(setq wff (car orig))
	(setq lit (cadr orig))
	(setq rank (getp wff 'RANK))
	(setq input-index (getp wff 'INPUT-INDEX))
	; fix - in case the wff is the same as a rule key
	(setq subs2 (remove-if #'(lambda (a) (equal (car a) wff)) subs2))
	(setq subs2 (append subs2 (list (list lit res))))
	(setq res (substitute-prop wff subs2))
	(setq prob (combine-probs prob *subst-prob*))
	(when (traced-p 'infer-details)
	      (print-line "    Input-driven inference: New inference will be " res))
	)

  (setq parents (union parents (list rule)))
  (setq prob (combine-probs prob (cond ((equal (setq p (getp rule 'prob)) 1)
					(list rule)) (t p))))

  ; send result to loading process - it will be normalized and checked there
  (setq *pieces* nil)
  (setq *ordered-pieces* nil)
  (cond ((null res) (return-from finish-infer nil))
	(*simplification-schemas* 
	 (list res (remove-if #'(lambda (a) (getp a 'spec-eval)) parents) prob))
	(t (setq nres 
	     (stuff-it res 
		 :parents (remove-if #'(lambda (a) (getp a 'spec-eval)) parents)
		 :prob prob :proof-prop proof-prop
		 :rank (+ rank 1) :input-index input-index
		 :assumption (if proof-prop t (if wff (getp wff 'assumption)))
		 :assumptions assumptions
		 :inherit (if wff (inherit-int wff (calc-prob-list prob)))
		 :rules-so-far (append rules-so-far (list rule))))
  (unless proof-prop (order-pieces))
  (setq *pieces* oldpieces)
  (setq *ordered-pieces* (append op *ordered-pieces*))
  nres)) )

(defun key-verify (lit &aux ans
		       (qa-iter *qa-iterations*))
  (setq *key-verify* t)
  (tweak '*qa-iterations* 10)
  ; how to use assumptions here?
  (setq ans (question (getp lit 'print) nil 2 "Key verification for " nil))
  (qa-cleanup nil)
  (tweak '*qa-iterations* qa-iter)
  (setq *key-verify* nil)
  (car ans))

(defun do-subs (list subs &aux res)
  (setq *interim* t)
  (setq res 
  (mapcar #'(lambda (a) 
               (list (normalize-wff (substitute-prop (car a) subs) (cadr a))
		     (cadr a) (caddr a))) 
	  list))
  (setq *interim* nil)
  res)

(defun ok-infer (wff1 wff2)
  (and (not (equal wff1 wff2))
       (if (getp wff1 'PROOF) (not (getp wff2 'DISPROOF)) t)
       (if (getp wff1 'DISPROOF) (not (getp wff2 'PROOF)) t)
       (if (getp wff2 'PROOF) (not (getp wff1 'DISPROOF)) t)
       (if (getp wff2 'DISPROOF) (not (getp wff1 'PROOF)) t)
       (right-type wff1 wff2)
       (not (getp wff1 'MEANING-POSTULATE))
       (not (getp wff2 'MEANING-POSTULATE))
       (not (getp wff1 'GOAL-ONLY))
       (not (getp wff2 'GOAL-ONLY))
       )
  )


(defun Nright-type (item wff)
  (and (or (getp item 'input-index) (getp wff 'input-index))
       (or (getp item 'conditional) (equal (getp item 'pred) 'or)
	   (getp wff 'conditional) (equal (getp wff 'pred) 'or))
       ))

(defun right-type (item wff)
  (cond ((or (getp item 'conditional) (equal (getp item 'pred) 'or))
	 (not (or (getp wff 'conditional) (equal (getp wff 'pred) 'or))))
	(t (or (getp wff 'conditional) (equal (getp wff 'pred) 'or))))
  )

(defun key-usefulness (lit &aux (use 0) (uselit (car lit)))
  ; if constant arg, start at 10
  (if (not (var-present uselit)) (setq use 10))
  ; if episodic add 5
  (if (episodic-lit uselit) (setq use (+ use 5)))
  ; subtract 1 for every unbound var
  (setq use (- use (list-length (find-vars uselit))))
  use)

(defun constant-present (lit &aux
			     (args (getp lit 'arguments)))
  (eval (append (list 'and) (mapcar #'(lambda (a) (not (epi-variable a))) args)))
  )

(defun order-keys (keys)
  ; each key current version, context and original
  (sort keys '> :key 'key-usefulness)
  )

(defun find-vars (lit &aux
		      varlist
		      (args (getp lit 'arguments)))
  (dolist (a args) 
	  (if (epi-variable a) 
	      (push a varlist)
	      (if (symbolp a)
		  (setq varlist (append varlist (find-vars a)))))
	  )
  varlist
  )

;;
;;    Meaning Postulate inferences
;; use above routines, but start a little differently (no "lit-classes" stuff)
;;

(defun mp-inferences (item &optional rules-so-far assumptions pr-prop &aux
			   (+keys (getp item '+keys))
			   (+other (getp item '+other))
			   results used-classes already-wff wffs
			   (proof-prop (cond (pr-prop)
					     (t (if (getp item 'PROOF) 'PROOF
					   (if (getp item 'DISPROOF)
					       'DISPROOF)))))
			   classes
			   )
  (unless nil 
	  (setq used-classes nil)
	  (setq already-wff nil)
	  (dolist (k (list item)) 
	          (setq classes (mp-classify k))
		  (do ((c (car classes) (car classes)))
		      ((null c))
		      (setq classes (cdr classes))
		      (unless (member c used-classes :test 'equal)
			      (push c used-classes)
			      (if (traced-p 'mp-class)
				  (print-line "Looking for mps for " 
					      (getp k 'print) 
					      " under class " c))
			      ; lookup mps
			      (setq wffs (get-class c))
			      (dolist (w wffs)
				      (when (and (getp w 'meaning-postulate)
						 (if *simplification-schemas*
						     (getp w 'simplification-schema)
						     (not (getp w 'simplification-schema)))
						 ;;(if *simplification-schemas*
						 ;;(not (member w rules-so-far))
						 ;;t)
						 ;;ANK above is orig
						 (not (member w rules-so-far))
						 (not (member w already-wff
							      :test 'equal)))
					    (push w already-wff)
					    (setq results (append results 
						  (try-with-wff t item k '+ w 
						     +keys nil +other nil 
						     proof-prop rules-so-far
						     assumptions)))
					    )
				      (when (and *simplification-schemas* results) 
					    (return))
				      ))
		      (when (and *simplification-schemas* results) (return))
		      ;; add classifications for parents of these now
		      (setq classes (union classes (super-class c t) 
					   :test 'equal))
		      )
		  (when (and *simplification-schemas* results) (return))
		  ))
  results)

(defun add-for-tracing (&optional item)
  (cond ((null item) nil)
	((listp item) 
	 (setq *minimum-trace-items* (append *minimum-trace-items* (list (car item))))
	 item)
	(t (setq *minimum-trace-items* (append *minimum-trace-items* (list item)))
	   item))
  )
