;       Last change    August 1993
(in-package epilog)

;; Question-Answering routines

;; When a question is first recieved, an attempt is made to verify it and
;; its negation. If either of these gives an answer, we stop.
;; The next step is to start the proof and disproof attempts by taking the
;; questioned wff, and its negation, and determining the goal-directed access
;; clssifications which may lead to an answer. Tese access actions are 
;; added to the agenda.
;; Items are taken from the top of the agenda, and handled, until one of the
;; following stopping criteria is met:
;;     maximum number of iterations (*qa-iterations*) has been reached
;;     no more items on the agenda (answer unknown)
;;     an answer is determined whose probability is above *question-threshold*
;; For access actions:
;;    The relevant wffs are retrieved, for a new access.
;;    For a continued access, the wffs not handled yet are used.
;;    Each wff is compared against the set-of-support wff. If there is a
;;    match, subgoal actions are added to the agenda, and this access action
;;    is added back to the agenda (next time it will be a continued access)
;; For subgoal actions:
;;    The new subgoal is determined given the wffs, which literals matched,
;;    and substitution information in the action.
;;    This subgoal is verified - if it results in NO, we have an answer.
;;    If simplified, the simplified version is used.
;;    Access actions are then added for the new subgoal.

;  Useful routines:
;     (question wff)
;        starts the qa process off. If wff is not specified, it continues
;        working on the previous question. May be abbreviated to (q wff)
;     (print-agenda number)
;        prints number items from the top of the agenda, in readable form.
;        If number is ommitted, all items are printed.

;  Useful global variables and constants:

(tweakable '*minimum-effort* 6 
  "Minimum effort to expend in searching for alternate answers - # iterations"
  '(integer 0 100)) 
(tweakable '*say-answer* t 
  "Indicates whether to say answer to question in English" 'boolean)
(tweakable '*say-question* nil 
  "Indicates whether to say question itself in English" 'boolean)

(tweakable '*question-effort* 3 
  "Effort to use in answering q's 0=lookup/verify 1=allow subgoal splitting
     2=allow inference 3=allow assumptions" '(integer 0 3))
(tweakable '*qa-iterations* 40 ;75 ;10
  "Maximum number of iterations through the agenda
	for question answering" '(integer 0 1000))
(tweakable '*qa-depth* 10 ; 20 
  "Maximum depth to go while question answering" '(integer 0 100))
(tweakable '*max-wh-difference* 0
  "Maximum difference in depth allowed between first wh answer and others")
  ;'(integer 0 100) may also be nil
(tweakable '*question-threshold* 0.4
  "Probability threshold for accepting an answer
	Minimum probability an answer must have to be accepted"
   '(float 0.0 1.0))
(tweakable '*qa-access-weight* 40
  "Extra weight given to accesses over subgoals
      This determines how much subgoal actions are prefered over access actions"
  '(float * *))
(tweakable '*contra-weight* 600 ; 200 ;20 ;10
  "Extra weight given to contrapositive subgoals over regular subgoals"
  '(float * *))
(tweakable '*rank-importance* 100
  "Weight of rank in calculating agenda position
  Determines how much higher ranked actions (deeper) are prefered"
  '(float * *))
(tweakable '*prob-importance* 100
  "Weight of probability in calculating agenda position          
  Determines how much actions with higher probabilities are prefered"
  '(float * *))
(tweakable '*initial-prob* .75 
  "Probability to use in initial calculations when probability is not known"
  '(float 0.0 1.0))
(tweakable '*interest-importance* 5 ;10
  "Weight of interest in calculating agenda position          
  Determines how much actions with higher interest are prefered"
  '(float * *))
(tweakable '*complexity-importance* 20 ;5 
  "Weight of complexity in calculating agenda position              
  Determines how much actions with lower complexity are prefered"
  '(float * *))
(tweakable '*difficulty-importance* 50 
  "Weight of difficulty in calculating agenda position              
  Determines how much actions with lower difficulty are prefered"
  '(float * *))
(tweakable '*goal-forward* nil
  "Indicates whether ok to try forward inference in question answering"
  'boolean)
(tweakable '*save-results* nil
  "Indicates whether or not to save results from queries"
  'boolean)
; if #rules here 0, specialist evaluation
; 0 saves all results that required inference, 1 saves those that had at least
; one rule apply, 2 for those that had at least 2 rules apply, etc.
; (problem - does not take into account the number of times a rule applied)
(tweakable '*result-difficulty* 0
  "Indicates how difficult the answer had to be to save it - the number of rules applied"
  '(integer 0 100))
(tweakable '*result-forward* nil
  "Indicates whether to allow forward inference when entering query results"
  'boolean)
(tweakable '*use-inherit* t
  "Indicates whether or not to use inherited interest in agenda positioning"
  'boolean)
(tweakable '*residue-penalty* 10000
  "Penalty for comparisons which involve a residue" '(float * *))
(tweakable '*mp-weight* 10000
  "Amount to lower access actions for mp's by (less preferred than regular)"
  '(float * *))
(tweakable '*finished-weight* 8000
  "Amount to lower access actions which have a solution (non wh)"
  '(float * *))
(tweakable '*var-for-var-penalty* 2000
  "How much to penalize subgoals that match variables with variables"
  '(float * *))
(tweakable '*favor-position* 100 
  "Amount higher in agenda to place the selected subgoal"
  '(float * *))
(tweakable '*favor-interest* 1000
  "Amount to prime inherited interest of new subgoal with to keep it going for a while"
  '(float * *))

(traceable 'qa-test  "Shows tests for incompatibility between set-of 
	      support clauses and accessed clauses")
(traceable 'qa-success "Shows successful subgoal actions during the qa process")
(traceable 'qa-access "Shows access actions done during the qa process")
(traceable 'qa-eval "Shows evaluations and simplifications that take place
	      during the qa process")
(traceable 'qa-answer "Traces output of the answer")
(traceable 'qa-time "Keeps track of how much time each question takes")
(traceable 'qa-iterations 
  "Keeps track of how many iterations each question takes")
(traceable 'qa "Traces the minimum essentials for question answering"
     '(qa-time qa-iterations qa-answer))
(traceable 'qa-int "Traces the minimum required to see what's happening"
     '(qa-time qa-iterations qa-answer qa-success subgoal))
(traceable 'qa-all "Traces everything during qa"
     '(qa-time qa-success qa-access qa-eval qa-test subgoal qa-iterations 
	       qa-answer))
(traceable 'qa-after "Indicates subgoal path should be printed after the fact")

(defvar *old-qa-effort* nil "Temporary storage for question-effort")
(defvar *qa-mode* nil "true when answering a question")
(defvar *qa-start-time* 0 "start time of answering question")
(defvar *qa-number-iterations* 0 "number of iterations used for current question")
(defvar *continue-til-minimum* nil 
  "flag indicating how far to go looking for another answer") 
(defvar *qa-start-real-time* 0 "start in real time of answering question")
(defvar *question-answer* 'no-question "Answer to question or state of qa")
(defvar *question* nil "The current question being asked")
(defvar *normed-question* nil "The normalized current question being asked")
(defvar *solutions* nil "Keeps track of all answers obtained and probs")
(defvar *other-answers* nil
  "Keeps track of the answers with probability less then *question-threshold*")
(defvar *number-answers* nil "Number of answers to stop after")
(defvar *wh-pattern* nil "List of wh quantified variables in current question")
(defvar *e-pattern* nil "List of e quantified variables in current question")
(defvar *matches* nil "Keeps track of all matches for wh questions")
(defvar *e-matches* nil "Keeps track of all matches for wh questions")
(defvar *qa-agenda* nil "Agenda of access and subgoal items")
(defvar *max-depth* 0 "Maximum depth reached for any of solutions")
(defvar *max-wh-depth* nil "Depth used for first wh answer")
(defvar *qa-only-try* nil "Indicates proof/disproof if only one should be tried")
(defvar *ok-to-rearrange* nil 
  "Indicates whether conditions ok for rearranging agenda if necessary")
(defvar *current-attempt* nil
  "Indicates whether current attmept is for proof or disproof for rearrange")

(defstruct subgoal
  (parent nil) 			; parent goal (if nil, this is the top)
  (proof-prop nil)		; indicates whether proof or disproof
  (subgoal nil)			; subgoal wff (normalized)
  (complexity nil)		; subgoal complexity
  (interest nil)		; subgoal interest
  (inherit-interest nil)	; interest inherited from parent
  (prob nil)			; probability
  (c-prob nil)			; calculated probability
  (assumptions nil)		; assumptions up to here
  (answer nil)			; answer when obtained
  (combo nil)			; how to combine subgoal answers
  (subgoals nil)		; subgoals beneath this one
  (depth nil)			; depth this subgoal is at
  (difficulty 0)		; extra amount added for splitting
  (rules nil)			; rules used and which keys were used
  (disallowed-rules nil)	; rules used- do not apply immediately
  (wffs nil)			; wffs involved in answer so far
  (subs nil)			; for use in wh questions
  (agenda nil)			; list of agenda subgoal actions
  (previous nil)		; previous subgoals in chain
  (useful nil)			; indicates if actually in main proof/disproof
  (access-items nil)		; access items for this particular subgoal
  )

(defstruct (agenda-item (:print-object
			 (lambda (o s)
			   (format s "~a~%"
				   (list "ACT-TYPE:" (agenda-item-action-type o)
					 "POS:" (agenda-item-position o)
					 "SUBG-ACT:" (agenda-item-subgoal-action o)
					 "ACC-ACT:" (agenda-item-access-action o)
					 "COMP:" (agenda-item-complexity o)
					 "INH-INT:" (agenda-item-inherit-interest o)
					 "P:" (agenda-item-prob o)
					 "CP:" (agenda-item-c-prob o)
					 "RUL:" (agenda-item-rules o)
					 "DIS-RUL:" (agenda-item-disallowed-rules o)
					 )
				   )
			   )))
  (action-type nil)		; access, subgoal
  (subgoal-parent nil)		; subgoal item that is the parent of this
  (position nil)		; number to use in ordering agenda
  (subgoal-action nil)		; comparisons to be used
  (access-action nil)		; access info
  (complexity nil)		; action complexity
  (interest nil)		; action interest
  (inherit-interest nil)	; interest inherited from parent
  (prob nil)			; prob including wff compared with
  (c-prob nil)			; calculated prob
  (rules nil)			; rules used and which keys were used
  (disallowed-rules nil)	; rules used- do not apply immediately
  )

(defstruct access-info
  (literal nil)			; literal to classify and compare with
  (context nil)			; context of that literal
  (classes nil)			; classes yet to check for lit-in-subgoal
  (interest 0)			; interest of literal
  (inherit-interest nil)	; interest inherited from parent
  (current-class nil)		; current classification
  (wff-list nil)		; list of wffs from that access
  (classes-tried nil)  		; classes tried already
  (wffs-tried nil)		; wffs tried already
  (flags nil)			; flags indicating -sub or -super
  (mp nil)			; indicates mp class
  )

(defstruct subgoal-info
  (key nil)			; key compared from subgoal
  (context nil)			; context of key
  (kn-wff nil)			; wff obtained from kb to compare with
  (kn-wff-key nil)		; key from kn-wff
  (kn-wff-context nil)		; context of kn-wff key
  (comparison nil)		; comparison
  )



;; this literal-usefulness calculation is used for ordering access
;; actions on the agenda for proof by contradiction.  determining the
;; next key to use for forward inferencing, and determining the
;; literal to backchain on in backchaining

(defun literal-usefulness (lit)
  (+ (wff-complexity lit)
     (if (equal (prop-pred lit) '_)
         -10
     (if (or (episodic-lit lit)
             (useful-nonepisodic-lit lit))
         1
         0)) )
  )

(defun reorder-agenda ()
  ; go through agenda and recalculate all positions
  (dolist (a *qa-agenda*)
	  (setf (agenda-item-position a) (calc-position a)))
  ; resort agenda based on new numbers
  (setq *qa-agenda* (sort *qa-agenda* '< :key 'agenda-item-position))
  )

; calc-position determines a number to use in the positioning of items on
; the agenda, based on the type of action, the complexity of clauses
; involved, and the probability

(defun calc-position (agenda &aux
			     (subgoal (agenda-item-subgoal-parent agenda))
			     (type (agenda-item-action-type agenda))
			     (complexity (agenda-item-complexity agenda))
			     (depth (subgoal-depth subgoal))
			     (interest (max (agenda-item-interest agenda)
					  (if *use-inherit*
				       (cond ((agenda-item-inherit-interest agenda)) (t 0)) 0)))
			     (difficulty (subgoal-difficulty subgoal))
			     (prob (agenda-item-c-prob agenda))
			     (half (round (/ *qa-depth* 2)))
			     action pos comp)
  (when (and *max-wh-depth*)
        (if (> depth (if (equal type 'access) (- *max-wh-depth* 1) *max-wh-depth*))
	    (setq depth (- depth (cond (*qa-depth*) (t 10))))) )
  (setq pos 
  (case type
        (access
	   (setq action (agenda-item-access-action agenda))
           (+ *qa-access-weight*
	      (if (neg-p (access-info-literal action)) 
		  *contra-weight* 0)
	      (if (access-info-mp action) *mp-weight* 0)
              (* -1 (literal-usefulness (access-info-literal action)))
	      (* (if (> depth half) (- *qa-depth* depth) depth) -1
		 *rank-importance*)
	      (* -1 (/ interest (if (> complexity 0) complexity 1)) 
		 *interest-importance*)
              (* difficulty *difficulty-importance*)
              (* (if (equal depth 0) *initial-prob* prob) -1 *prob-importance*))
	 )
        (subgoal 
	   (setq action (agenda-item-subgoal-action agenda))
	   (setq comp (subgoal-info-comparison action))
	   (+ (if (equal (subgoal-info-kn-wff-context action) '-)
		  *contra-weight* 0)
	      (if (or (comparison-info-residue1 comp)
		      (comparison-info-residue2 comp))
		  *residue-penalty* 0)
	       (* (if (> depth half) (- *qa-depth* depth) depth) -1
		 *rank-importance*)
	      (if (or (var-for-var (comparison-info-subs1 comp))
		      (var-for-var (comparison-info-subs2 comp)))
		  *var-for-var-penalty* 0)
	      (* -1 (/ interest (if (> complexity 0) complexity 1)) 
		 *interest-importance*)
	      (* difficulty *difficulty-importance*)
              (* (if (equal depth 0) *initial-prob* prob) -1 *prob-importance*))
	 )
        (otherwise 10000)))
  (setf (agenda-item-position agenda) pos) 
  pos)

(defun var-for-var (sublist)
  (member-if #'(lambda (a) (epi-variable-p (cadr a))) sublist))

; add-to-agenda adds an item to the relevant place in the agenda
   
(defun add-to-agenda (item resta &aux p1 p2)
  (cond ((null resta) (list item))
        ((< (setq p1 (agenda-item-position item))
            (setq p2 (agenda-item-position (car resta))))
         (append (list item) resta))
        ; prefer subgoals to accesses
        ((and (= p1 p2)
              (equal (agenda-item-action-type item) 'access)
	      (equal (agenda-item-action-type (car resta)) 'subgoal))
         (append (list (car resta)) (add-to-agenda item (cdr resta))))
        ; new items preferred to old
        ((= p1 p2) (append (list item) resta))
        (t (append (list (car resta)) (add-to-agenda item (cdr resta))))
        )
  )

; add-agenda adds an item to the agenda
 
(defun add-agenda (item &optional (calc t))
  (when calc (calc-position item))
  (setq *qa-agenda* (add-to-agenda item *qa-agenda*))
  )
 
; agenda-search goes through the agenda, doing access and subgoal actions,
; until one of the stopping criteria are met.
 
(defun agenda-search (&aux item curpos done-wh
			   (teststop (/ *qa-iterations* 2)))
  (do ((i 0 (+ i 1)))
      ;; stopping conditions:
      ((or 
	(>= (setq curpos i) *qa-iterations*)
	(and *continue-til-minimum* 
	     (>= *qa-number-iterations* *continue-til-minimum*))
					; greater than # iterations allowed
	(null *qa-agenda*)		; nothing on agenda
	(if (and (> i 0) *wh-pattern* *max-wh-depth*)
	    (let* ((ni (car *qa-agenda*))
		   (p (agenda-item-subgoal-parent ni))
		   (d (subgoal-depth p)))
	      (if (if (equal (agenda-item-action-type ni) 'access)
		      (>= d *max-wh-depth*)
		    (> d *max-wh-depth*))
		  (setq done-wh t))))
	))
    (when (and *ok-to-rearrange* (not *wh-pattern*) (> *qa-iterations* 5)
	       (> *qa-number-iterations* teststop))
      (setq *ok-to-rearrange* nil)
      (rearrange-agenda *current-attempt*) )

    (setq *qa-number-iterations* (+ *qa-number-iterations* 1))
    (setq item (pop *qa-agenda*))

    (if *ok-to-rearrange*
	(if (and *current-attempt*
		 (not (equal *current-attempt* 
			     (subgoal-proof-prop (agenda-item-subgoal-parent item)))))
	    (setq *ok-to-rearrange* nil)
	  (setq *current-attempt* 
	    (subgoal-proof-prop (agenda-item-subgoal-parent item)))) )

    (case (agenda-item-action-type item)
      (access (access item))
      (subgoal (subgoal item)))
    )

  (if done-wh (print-line "Depth much lower than previous answers")
    (unless (or *exists-q* *key-verify* *insertion*
		(and *wh-pattern* *solutions*))
      (if (null *qa-agenda*) (print-line "No more actions on agenda"))
      (if (>= curpos *qa-iterations*)
	  (print-line "Maximum number of iterations (" *qa-iterations*
		      ") reached. Use (q) to continue.")) ) )
  )

(defun rearrange-agenda (cur-att &aux p
				 (lookfor (if (equal cur-att 'PROOF) 'DISPROOF
					      'PROOF)))
  (when (traced-p 'qa-success)
	(print-line "Spent too long trying " cur-att 
		    " attempt - going to try the " lookfor " attempt now"))
  (dolist (a *qa-agenda*)
	  (when (equal (subgoal-proof-prop (agenda-item-subgoal-parent a)) 
		       lookfor)
		; may have to change it's position number?
		(setf (agenda-item-position a)
		      (- (agenda-item-position (car *qa-agenda*)) 
			 *favor-position*))
		(setq p (agenda-item-subgoal-parent a))
		(setf (subgoal-inherit-interest p)
		      (max (subgoal-inherit-interest p) *favor-interest*))
		(setf (agenda-item-inherit-interest a) 
		      (subgoal-inherit-interest p))
		(setq *qa-agenda* (append (list a) (remove a *qa-agenda*)))
		(return)
		))
  )

(defvar *wff-subgoals* nil)

;;; Qa-cleanup cleans up the net after answering a question.
;;;
(defun qa-cleanup (&optional (reset t))
  (when reset (reset-indent "" 0))
  (when (not (eq *question-answer* 'no-question))
	(if (traced-p 'qa-test) 
	    (print-line  "Cleaning up after the previous question."))
	)
  (rewind-question)
  (dolist (w *wff-subgoals*) (remprop w 'subgoals))
  (setq *wff-subgoals* nil)
  (when *old-qa-effort*
	(setq *question-effort* *old-qa-effort*)
	(setq *old-qa-effort* nil))
  (setq *question-answer* 'no-question)
  (setq *qa-agenda* nil)
  (setq *qa-mode* nil)
  (setq *new-vars* t)
  (setq *e-pattern* nil)
  (setq *wh-pattern* nil)
  (setq *qa-only-try* nil)
  (setq *current-attempt* nil)
  (setq *simplification-schemas* nil)
  (setq *ok-to-rearrange* t)
  (when reset (temp-cleanup))
  (setq *qa-number-iterations* 0)
  (setq *max-depth* 0)
  (setq *max-wh-depth* nil)
  (setq *qa-start-time* (get-internal-run-time))
  (setq *qa-start-real-time* (get-internal-real-time))
  )


(defun proof-q (&optional qphi number effort printp (top t))
  (question qphi number effort printp top 'proof))

(defun pq (&optional qphi number effort printp (top t))
  (question qphi number effort printp top 'proof))

(defun disproof-q (&optional qphi number effort printp (top t))
  (question qphi number effort printp top 'disproof))

(defun dq (&optional qphi number effort printp (top t))
  (question qphi number effort printp top 'disproof))

;; question attempts to determine the truthfulness of phi.
(defun question (&optional qphi number effort printp (top t) only
		 &aux phi ans sol sublist newi)
  (setq *continue-til-minimum* nil)
  (if (equal qphi 'epilog::*question*) (setq qphi *question*))
  (catch 'question-catcher 
    (if qphi 
	;; new question
	(let ()
	  ;; cleanup after any old question, and set things up for this one
	  (qa-cleanup top)
	  (setq *qa-only-try* only)
	  (if (or *qa-only-try* *key-verify*) (setq *ok-to-rearrange* nil))
	  (setq *solutions* nil)
	  (setq *other-answers* nil)
	  (setq *matches* nil)
	  (setq *e-matches* nil)
	  (setq *proof* nil)
	  (setq *disproof* nil)
	  (setq *question* nil)
	  (when (and effort (not (= effort *question-effort*)))
	    (setq *old-qa-effort* *question-effort*)
	    (tweak '*question-effort* effort))
	  (setq *question-answer* 'inferencing)
	  (setq *number-answers* number)
	  (when (> *question-effort* 2)	;only necessary if assumptions -almost
	    (checkpoint-question))
	  (setq phi (normalize qphi))
	  (unless phi (return-from question nil))
	  (set-wh-pattern phi) 
	  (if *save-results* (set-e-pattern phi) )
	  (setq *question* qphi)
	  (ask-question phi printp)
	  )
      ;; continue on from where last question attempt left off
      (when *question*
	(if (> *question-effort* 0)
	    (if (and (not (equal *question* 'no-question))
		     *solutions* (null *qa-agenda*))
		;; have to initialize agenda first
		(try-harder (normalize *question*))
	      (agenda-search))) )))
  (say-results)
  (when (and (not *key-verify*) (not *insertion*) (not *exists-q*)
	     (traced-p 'qa-after)) 
    (print-subgoal))
  (setq ans *solutions*)
  (when (and *save-results* *solutions*
	     (not *exists-q*) (not *key-verify*) (not *insertion*))
    (cond ((or *wh-pattern* *e-pattern*)
	   (let ((vars *wh-pattern*)
		 (evars *e-pattern*)
		 (sol *solutions*)
		 (emtch *e-matches*)
		 (qq (cond (*normed-question*) 
			   (t (normalize *question* '-)))))
	     ;; take each solution, make substitutions, save
	     (do ((s (car sol) (car sol))
		  (em (car emtch) (car emtch)))
		 ((null s))
	       (setq sol (cdr sol))
	       (setq emtch (cdr emtch))
	       (setq sublist 
		 (append (if vars (make-wh-subs (car s) vars))
			 (if evars (make-wh-subs em evars))))
	       (setq newi (substitute-prop qq sublist))
	       (if (equal (car s) 'NO)
		   (setq newi (negate newi)))
	       (stuff-result newi (cadr s) (caddr s))
	       )))
	  ((setq sol (combine-answers))
	   (cond ((equal (car sol) 'YES)
		  (stuff-result *question* (cadr sol) (caddr sol)))
		 ((equal (caar *solutions*) 'NO)
		  (stuff-result (negate *question*) (cadr sol) (caddr sol)))
		 ))
	  ))
  ans)

(defun make-wh-subs (items vars &aux sublist)
  (do ((i (car items) (car items))
       (v (car vars) (car vars)))
      ((or (null i) (null v)))
      (setq items (cdr items))
      (setq vars (cdr vars))
      (setq sublist (append sublist (list (list v i))))
      )
  sublist)


(defun stuff-result (phi parents prob &aux res
			 (rules (remove-if-not 
				 #'(lambda (a) (getp a 'conditional)) parents))
			 (rf *rule-forward*)
			 (sf *story-forward*))
  (when (or (>= (list-length rules) *result-difficulty*)
	    (and (equal *result-difficulty* 0)
		 (member-if #'(lambda (a) (getp a 'spec-eval)) parents)) )
	(qa-cleanup t)
	(unless *result-forward*
		(setq *story-forward* nil)
		(setq *rule-forward* nil)
		)
	(print-line "Asserting answer to previous query: " phi)
	(setq res (stuff-it phi :prob prob 
	        :parents (remove-if #'(lambda (a) (getp a 'spec-eval)) parents)
		:rank (if rules (list-length rules) 1)
		:rules-so-far rules))
	(setq *story-forward* sf)
	(setq *rule-forward* rf) )
  res)

(defun set-wh-pattern (phi)
  (setq *wh-pattern* (getp phi 'wh-vars)) )
(defun set-e-pattern (phi)
  (setq *e-pattern* (getp phi 'e-vars)) )
 
; q is a short form for question

(defun q (&optional phi number effort printp (top t) only)
  (question phi number effort printp top only))

; ask-question attempts to answer a new question
; First, it attempts to verify phi and its negation; if this fails, it
; starts up the agenda, and proceeds to do goal-directed reasoning.

(defun ask-question (phi printp &aux first-attempt)
  (if printp
      (if (traced-p 'forward-qa)
	  (print-line printp " " (getp phi 'print)))
    (progn
      (print-line "Questioning" " " (getp phi 'print))
      (when *say-question* (say-it (list phi)))
      )
    )
  (setq *qa-mode* t)
  (setq first-attempt (verify phi nil))
  (when (and first-attempt (member (car first-attempt) '(YES NO)))
    (if (and (solution first-attempt)
	     *qa-mode*
	     (or (not *wh-pattern*) 
		 (and *number-answers* (<= *number-answers* 1)
		      (let () (print-line "Stopping after " *number-answers*
					  " answers") t))))
	(throw 'question-catcher t)) )
  (when (and first-attempt (not (member (car first-attempt) '(YES NO))))
    (when (or (traced-p 'qa-success) (traced-p 'qa-eval))
      (print-line "Simplified question " (getp phi 'print) " to "
		  (getp (car first-attempt) 'print)))
    (setq phi (car first-attempt))
    )

  ;; if we actually get here, either the question hasn't been answered
  ;; yet or we have a wh-question looking for all possible answers in
  ;; 2nd case, if some answers have been found, stop
  (if (or (= *question-effort* 0) (and *wh-pattern* *solutions*)) 
      (throw 'question-catcher nil))
  (try-harder phi)
  )

(defun assumption (item) (getp item 'assumption))

;; solution takes a given solution, extracts wh-q matches from it if
;; necessary, adds it to the list of answers, and stops the qa process
;; if the required number of answers have been obtained.

;; NOTE found a discrepancy between the documentation and the
;; code. Thanks to an email from Alexander S. Yeh. The documentation
;; talk about a global variable *other-answers* that is supposed to
;; save the answer with a probability below the threshold defined in
;; *question-threshold* but in the all code there is no trace of this
;; variable. For the moment we have chosen to don't use the answer
;; with probability less than *question-threshold*. FM - Jan 25 2005.
;; Moreover the part of code managing the *question-threshold* was
;; only printing a warning but not taking the action it was supposed
;; to take (i.e. in case the answer was below the threshold keep
;; searching for other answers). FM - Jan 25 2005.

(defun solution (result &optional subs (depth 0) &aux p newsol
						      wffs usewffs prob useprob ematch match already)
  (setq *max-depth* (max *max-depth* depth))
  ;; if wh question, add solution to list and continue if max number
  ;; of answers reached or yes/no question, return to top - done
  (when (equal (car result) 'YES)
    (cond (*wh-pattern* 
	   (unless (or *max-wh-depth* (not *max-wh-difference*))
	     (setq *max-wh-depth* (+ depth *max-wh-difference*))
	     (reorder-agenda))
	   (setq match (make-match *wh-pattern* subs))
	   (when (member match *matches* :test 'equal)
	     (if (traced-p 'qa-success)
		 (print-line "Solution " match " already found "))
	     (return-from solution nil))
	   (push match *matches*) 
	   (setq ematch (make-match *e-pattern* subs))
	   (push ematch *e-matches*))
	  (*e-pattern*
	   (setq ematch (make-match *e-pattern* subs))
	   (push ematch *e-matches*))
	  ))
  (setq wffs (cadr result))
  (setq already wffs)
  (do ((w (car wffs) (car wffs)))
      ((null w))
    (setq wffs (cdr wffs))
    (if (getp w 'assumption)
	;; ANK modified this to handle cycles in the parentage
	;; graph correctly.  I don't know if that should happen, but
	;; it does.  (6/20/00 I fixed the problem that made cycles
	;; happen, but I'll leave this the way it is just in case.)
	(dolist (w1 (getp w 'parents))
	  (when (not (member w1 already))
	    (setq wffs (cons w1 wffs))
	    (setq already (cons w1 already))))
      (setq usewffs (union usewffs (list w))))
    )
  (setq prob (caddr result))
  (setq already prob)
  (if (listp prob)
      (do ((w (car prob) (car prob)))
	  ((null w))
	(setq prob (cdr prob))
	(if (and (symbolp w) (getp w 'assumption))
	    (if (listp (getp w 'prob))
		;; ANK see note above
		(dolist (p1 (getp w 'prob))
		  (when (not (member p1 already))
		    (setq prob (cons p1 prob))
		    (setq already (cons p1 already))))
	      (setq useprob (combine-probs useprob (getp w 'prob))))
	  (setq useprob (combine-probs useprob (list w))))
	)
    (setq useprob prob))
  (setq newsol (list (if match match (car result))
		     usewffs useprob))
  (if (and (not *wh-pattern*)
	   (<= (setq p (calc-prob-list (caddr result))) *question-threshold*))
      (progn
	(print-line "Answer below probability threshold: "
		    (car result) " with probability " p
		    *nl* *spaces* " Continuing search for better answer ... ")
	(pushnew newsol *other-answers* :test 
		 #'(lambda (a b) (not (not-same-path (cadr a) (cadr b)))))
	nil
	)
    (progn
      (when (traced-p 'qa-success)
	(print-line "** Answer: " (car newsol) " with probability " 
		    (calc-prob-list (caddr newsol)))
	)
      (pushnew newsol *solutions* :test 
	       #'(lambda (a b) (not (not-same-path (cadr a) (cadr b)))))
      t
      )
    )
  )

(defun make-match (vars subs &aux sub match)
  (dolist (v vars)
	  (setq sub (cadr (assoc v subs :test 'equal)))
	  (if (and (symbolp sub) (equal (getp sub 'term-type) 'function))
	      (setq sub (cond ((spec-eval-fn sub)) (t sub))))
	  (setq match (append match (list sub)))
	  )
  match)

(defun pr-solutions (brief &rest args) 
  (declare (ignore brief args))
  (say-results))

(set-display-function 'solutions 'pr-solutions
		      "Repeats the answers to the last question")
(set-display-function 'solution 'pr-solutions
		      "Repeats the answers to the last question")
(set-display-function 'answers 'pr-solutions
		      "Repeats the answers to the last question")
(set-display-function 'answer 'pr-solutions
		      "Repeats the answers to the last question")

(defun pr-question (brief &rest args)
  (declare (ignore brief args))
  (print-it *nl* "The current question is " *question*))
(set-display-function 'question 'pr-question "Displays the current question")
(set-display-function 'q 'pr-question "Displays the current question")

(defun say-results (&aux ans)
  (when (if (or *exists-q* *key-verify*) (traced-p 'forward-qa)
	  (traced-p 'qa-answer))
    (cond (*wh-pattern*
	   (print-line "Answer(s): ")
	   (if (null *solutions*) (print-it "UNKNOWN"))
	   (set-indent 5)
	   (dolist (s *solutions*) 
	     (print-line (if (listp (car s))
			     (mapcar 'print-info (car s)) 
			   (car s))
			 "   with probability " 
			 (calc-prob-list (caddr s))
			 )
	     )
	   (reset-indent "" 0)
	   (if (and *print* *say-answer*) 
	       (say-them (mapcar 'cadr *solutions*)))
	   )
	  ((setq ans (combine-answers))	; (car *solutions*)
	   (print-line "Answer: " (car ans) " with probability "
		       (calc-prob-list (caddr ans))
		       " (args of calc-prob-list: "
		       (caddr ans)
		       ")"
		       )
	   (if (and *print* *say-answer* (not (or *key-verify* *exists-q*)))
	       (respond ans))
	   )
	  (t (print-line "Answer: UNKNOWN"))
	  ))
  
  (unless (or *exists-q* *key-verify*)
	  (if (traced-p 'qa-time)
	      (print-line "    Used "
		  (/ (float (- (get-internal-run-time) *qa-start-time*))
		     internal-time-units-per-second) " seconds CPU time"
		  #\tab
		  (/ (float (- (get-internal-real-time) *qa-start-real-time*))
		     internal-time-units-per-second) " seconds real time"))
	  (when (traced-p 'qa-iterations)
	        (print-line "    Used " *qa-number-iterations* " iterations ")
		(if *solutions* (print-it "  Searched to depth " *max-depth*))))
  )

(defvar *proof* nil "Top level proof goal")
(defvar *disproof* nil "Top level proof goal")

;; try-harder sets up the agenda and begins the goal-directed reasoning
(defun try-harder (phi)
  ;; create proof and disproof wffs
  ;; start off agenda with access actions for each
  
  (unless (equal *qa-only-try* 'DISPROOF)
    (setq *proof* (add-subgoal nil phi :proof-prop 'PROOF)) )
  (unless (equal *qa-only-try* 'PROOF)
    (unless *exists-q*
      (setq *disproof*
	(add-subgoal nil (normalize (negate (getp phi 'print))) 
		     :proof-prop 'DISPROOF))) )
  
  ;; now go through agenda until stopping criteria reached
  (agenda-search)
  )

(defun pr-agenda (brief &rest args &aux (number (car args)))
  (declare (ignore brief))
  (print-agenda number))

(set-display-function 'agenda 'pr-agenda 
  "Prints out the question answering agenda")

; print-agenda prints items from the question answering agenda in readable form

(defun print-agenda (&optional number &aux cl comp
			       subgoal action l1 l2 subs1 subs2
			       (count 0))
  (dolist (i *qa-agenda*)
	  (setq subgoal (agenda-item-subgoal-parent i))
	  (setq count (+ count 1))
	  (when (and number (> count number))
		(return))
	  (print-line count ": " (agenda-item-action-type i) "  "
		      (if (equal (subgoal-proof-prop subgoal) 'PROOF) 
			  "PROOF" "DISPROOF") 
		      " Probability: " (agenda-item-prob i) " " 
		      (agenda-item-c-prob i)
		      "     Complexity: " (agenda-item-complexity i) 
		      "     Interest: " (agenda-item-interest i) 
		      "     Inherited Interest: " (agenda-item-inherit-interest i) 
		      *nl* *spaces*
		      "     Depth: " (subgoal-depth subgoal)
		      "     Position: " (agenda-item-position i)
		      "     Difficulty: " (subgoal-difficulty
					   (agenda-item-subgoal-parent i)))
	  (case (agenda-item-action-type i)
		(access (setq action (agenda-item-access-action i))
			(print-line "    " (subgoal-subgoal subgoal) ": " 
				    (getp (subgoal-subgoal subgoal) 'print) )
			(print-line "    using literal " 
				    (access-info-literal action) ": "
				    (getp (access-info-literal action)  'print)
				    " " (access-info-context action) " "
				    (if (access-info-mp action) "  (mp class)" 
					""))
			(if (access-info-current-class action)
			    (print-line "    Current class: "
					(access-info-current-class action)))
			(when (setq cl (access-info-classes action))
			      (print-line "    Classes: " (car cl))
			      (dolist (c (cdr cl))
				      (print-line "             " c))
			      ) 
			)
		(subgoal (setq action (agenda-item-subgoal-action i))
			 (setq comp (subgoal-info-comparison action))
			 (setq l1 (subgoal-info-key action))
			 (setq l2 (subgoal-info-kn-wff-key action))
			 (cond ((equal l1 (comparison-info-lit1 comp))
				(setq subs1 (comparison-info-subs1 comp))
				(setq subs2 (comparison-info-subs2 comp)) )
			       (t (setq subs1 (comparison-info-subs2 comp))
				  (setq subs2 (comparison-info-subs1 comp))))
			 (print-line "    "  "Wff to compare against: "
				     (subgoal-subgoal subgoal) ": "
				     (getp (subgoal-subgoal subgoal) 'print)) 
			 (print-line "    " "Wff retrieved: "
				     (subgoal-info-kn-wff action) ": "
				     (getp (subgoal-info-kn-wff action) 'print)) 
			 (print-line "    " "Literal in 1st wff: "
				     l1 ": " (print-info l1)
				     " " (subgoal-info-context action) )
			 (print-line "        with substitutions: " subs1)
			 (print-line "    " "Literal in 2nd wff: "
				     l2 ": " (print-info l2)
				     " " (subgoal-info-kn-wff-context action) )
			 (print-line "        with substitutions: " subs2)
			 (if (comparison-info-residue1 comp)
			     (print-line "    " "Residue1: "
				        (comparison-info-residue1 comp) ))
			 (if (comparison-info-residue2 comp)
			     (print-line "    " "Residue2: "
				        (comparison-info-residue2 comp)))
			 )
		(otherwise (print-line "    Illegal type: " ))
		))
  )

(defun reset (&optional clobber)
  (compare-cleanup)
  (class-cleanup)
  (qa-cleanup)
  (reset-indent "" 0)
  (when clobber (rewind-all))
  (setq *key-verify* nil)
  (setq *exists-q* nil)
  )


(defun combine-the-answers (solutions &aux answer prob s-wffs c-wffs temp)
  (if (null (cdr solutions)) (return-from combine-the-answers (car solutions)))
  (do ((s (car solutions) (car solutions)))
      ((null s))
      (setq solutions (cdr solutions))
      (cond ((and (null answer) (null prob))
	     (setq answer (car s))
	     (setq s-wffs (list (cadr s)))
	     (setq prob (calc-prob-list (caddr s))))
	    ((equal (car s) answer)
	     ; supporting
	     (setq s-wffs (append s-wffs (list (cadr s))))
	     (setq prob (- 1 (* (- 1 prob) (- 1 (calc-prob-list (caddr s)))))))
	    (t ; conflicting
             (let* ((newp (calc-prob-list (caddr s)))
		    (mult (* prob newp)))
		  (cond ((> newp prob)
			 ; new one wins
			 (setq answer (car s))
			 (setq prob (/ (- newp mult) (- 1 mult)))
			 (setq temp s-wffs)
			 (setq s-wffs c-wffs)
			 (setq c-wffs temp)
			 (setq s-wffs (append s-wffs (list (cadr s))))
			 )
			(t 
			 (setq c-wffs (append c-wffs (list (cadr s))))
			 (setq prob (/ (- prob mult) (- 1 mult))))
			))
	       ))
      )
  ; somehow list the wffs so that conflicting evidence is said with but??
  (list answer (car s-wffs) prob)
  )

(defun combine-answers (&aux (solutions *solutions*) yeslist nolist 
			     yesans noans)
  (dolist (s solutions)
    (if (equal (car s) 'YES)
	(setq yeslist (append yeslist (list s)))
      (setq nolist (append nolist (list s)))))
  (when yeslist (setq yesans 
		  (list 'YES (max-wff yeslist) 
			(combine-support-probs (mapcar 'caddr yeslist)))))
  (when nolist (setq noans 
		 (list 'NO (max-wff nolist) 
		       (combine-support-probs (mapcar 'caddr nolist)))))
  
  (if yesans
      (if noans
	  (if (> (caddr noans) (caddr yesans))
	      (list 'NO (cadr noans) 
		    (combine-contradict-probs (caddr noans) (caddr yesans)))
	    (list 'YES (cadr yesans) 
		  (combine-contradict-probs (caddr yesans) (caddr noans))) )
	yesans)
    noans)
  )

(defun max-wff (anslist &aux prob wlist)
  (do* ((ans (car anslist) (car anslist))
       (p (calc-prob-list (caddr ans)) (calc-prob-list (caddr ans))))
      ((null ans))
      (setq anslist (cdr anslist))
      (when (or (null prob) (> p prob))
	    (setq prob p)
	    (setq wlist (cadr ans)))
      )
  wlist)
